#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.

from octobot.api import backtesting
from octobot.api import strategy_optimizer
from octobot.api import updater

from octobot.api.backtesting import (
    create_independent_backtesting,
    check_independent_backtesting_remaining_objects,
    is_independent_backtesting_in_progress,
    is_independent_backtesting_computing,
    get_independent_backtesting_progress,
    is_independent_backtesting_finished,
    is_independent_backtesting_stopped,
    get_independent_backtesting_exchange_manager_ids,
    get_independent_backtesting_bot_id,
    log_independent_backtesting_report,
    get_independent_backtesting_config,
    get_independent_backtesting_symbols_by_exchanges,
    initialize_and_run_independent_backtesting,
    join_independent_backtesting,
    initialize_independent_backtesting_config,
    clear_backtesting_fetched_data,
    stop_independent_backtesting,
    join_independent_backtesting_stop,
    get_independent_backtesting_report,
)
from octobot.api.strategy_optimizer import (
    create_strategy_optimizer,
    create_design_strategy_optimizer,
    find_optimal_configuration,
    initialize_design_strategy_optimizer,
    update_strategy_optimizer_total_runs,
    generate_and_save_strategy_optimizer_runs,
    create_strategy_optimizer_settings,
    resume_design_strategy_optimizer,
    cancel_strategy_optimizer,
    print_optimizer_report,
    get_optimizer_report,
    get_optimizer_results,
    get_optimizer_overall_progress,
    get_design_strategy_optimizer_queue,
    update_design_strategy_optimizer_queue,
    is_optimizer_in_progress,
    is_optimizer_computing,
    is_optimizer_finished,
    get_optimizer_errors_description,
    get_optimizer_current_test_suite_progress,
    get_optimizer_strategy,
    get_optimizer_all_time_frames,
    get_optimizer_all_TAs,
    get_optimizer_all_risks,
    get_optimizer_trading_mode,
    get_optimizer_is_properly_initialized,
)
from octobot.api.updater import (
    get_updater,
)

__all__ = [
    "create_independent_backtesting",
    "check_independent_backtesting_remaining_objects",
    "is_independent_backtesting_in_progress",
    "is_independent_backtesting_computing",
    "get_independent_backtesting_progress",
    "is_independent_backtesting_finished",
    "is_independent_backtesting_stopped",
    "get_independent_backtesting_exchange_manager_ids",
    "get_independent_backtesting_bot_id",
    "log_independent_backtesting_report",
    "get_independent_backtesting_config",
    "get_independent_backtesting_symbols_by_exchanges",
    "initialize_and_run_independent_backtesting",
    "join_independent_backtesting",
    "initialize_independent_backtesting_config",
    "clear_backtesting_fetched_data",
    "stop_independent_backtesting",
    "join_independent_backtesting_stop",
    "get_independent_backtesting_report",
    "create_strategy_optimizer",
    "create_design_strategy_optimizer",
    "find_optimal_configuration",
    "initialize_design_strategy_optimizer",
    "update_strategy_optimizer_total_runs",
    "generate_and_save_strategy_optimizer_runs",
    "create_strategy_optimizer_settings",
    "resume_design_strategy_optimizer",
    "cancel_strategy_optimizer",
    "print_optimizer_report",
    "get_optimizer_report",
    "get_optimizer_results",
    "get_optimizer_overall_progress",
    "get_design_strategy_optimizer_queue",
    "update_design_strategy_optimizer_queue",
    "is_optimizer_in_progress",
    "is_optimizer_computing",
    "is_optimizer_finished",
    "get_optimizer_errors_description",
    "get_optimizer_current_test_suite_progress",
    "get_optimizer_strategy",
    "get_optimizer_all_time_frames",
    "get_optimizer_all_TAs",
    "get_optimizer_all_risks",
    "get_optimizer_trading_mode",
    "get_optimizer_is_properly_initialized",
    "get_updater",
]



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.

from octobot.community.models import community_user_account
from octobot.community.models.community_user_account import (
    CommunityUserAccount,
)
from octobot.community.models import community_fields
from octobot.community.models.community_fields import (
    CommunityFields,
)

from octobot.community.models import community_tentacles_package
from octobot.community.models import community_supports
from octobot.community.models import community_donation
from octobot.community.models import startup_info

from octobot.community.models.community_tentacles_package import (
    CommunityTentaclesPackage
)
from octobot.community.models.community_supports import (
    CommunitySupports
)
from octobot.community.models.community_donation import (
    CommunityDonation
)
from octobot.community.models.startup_info import (
    StartupInfo
)
from octobot.community.models.formatters import (
    format_trades,
    format_orders,
    format_portfolio,
    format_portfolio_history,
    format_portfolio_with_profitability,
)
from octobot.community.models.community_public_data import (
    CommunityPublicData
)

__all__ = [
    "CommunityUserAccount",
    "CommunityFields",
    "CommunityTentaclesPackage",
    "CommunitySupports",
    "CommunityDonation",
    "StartupInfo",
    "format_trades",
    "format_orders",
    "format_portfolio",
    "format_portfolio_history",
    "format_portfolio_with_profitability",
    "CommunityPublicData",
]



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import math

import pytest
import pytest_asyncio
import mock

import octobot.api as bot_module_api
import octobot.enums as enums
import octobot.strategy_optimizer as strategy_optimizer
import octobot_trading.api as trading_api
import octobot_tentacles_manager.api as tentacles_manager_api
import octobot_commons.enums as commons_enums
import octobot_commons.constants as commons_constants
import octobot_commons.databases as databases
import octobot_evaluators.constants as evaluators_constants

import tentacles.Evaluator.Strategies as Strategies
import tentacles.Evaluator.TA as Evaluator
import tentacles.Trading.Mode as Mode

from tests.unit_tests.strategy_optimizer import optimizer_inputs, MOCKED_OPTIMIZER_CONFIG, EXPECTED_RUNS_FROM_MOCK

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_generate_and_save_strategy_optimizer_runs(optimizer_inputs):
    tentacles_setup_config, trading_mode = optimizer_inputs
    optimizer_settings = bot_module_api.create_strategy_optimizer_settings({
        enums.OptimizerConfig.OPTIMIZER_CONFIG.value: MOCKED_OPTIMIZER_CONFIG,
    })
    with mock.patch.object(databases.TinyDBAdaptor, "create_identifier", mock.AsyncMock()) as create_identifier_mock, \
            mock.patch.object(strategy_optimizer.StrategyDesignOptimizer, "_save_run_schedule", mock.AsyncMock()) as \
                    _save_run_schedule_mock, \
            mock.patch.object(strategy_optimizer.StrategyDesignOptimizer, "shuffle_and_select_runs", mock.Mock(
                side_effect=lambda *args, **_: args[0]
            )) as shuffle_and_select_runs_mock:
        assert await bot_module_api.generate_and_save_strategy_optimizer_runs(
            trading_mode,
            tentacles_setup_config,
            optimizer_settings
        ) == EXPECTED_RUNS_FROM_MOCK
        create_identifier_mock.assert_called_once()
        shuffle_and_select_runs_mock.assert_called_once()
        _save_run_schedule_mock.assert_awaited_once_with(EXPECTED_RUNS_FROM_MOCK)


async def test_resume_unknown_mode(optimizer_inputs):
    tentacles_setup_config, trading_mode = optimizer_inputs
    optimizer_settings = bot_module_api.create_strategy_optimizer_settings({
        enums.OptimizerConfig.OPTIMIZER_CONFIG.value: MOCKED_OPTIMIZER_CONFIG,
        enums.OptimizerConfig.OPTIMIZER_IDS.value: [1],
        enums.OptimizerConfig.MODE.value: "plop",
    })
    optimizer = bot_module_api.create_design_strategy_optimizer(
        trading_mode,
        optimizer_settings,
        None,
        tentacles_setup_config,
    )
    with mock.patch.object(optimizer, "_get_total_nb_runs",
                           mock.AsyncMock(return_value=5)) as _get_total_nb_runs_mock:
        with pytest.raises(NotImplementedError):
            await optimizer.resume(optimizer_settings)
        _get_total_nb_runs_mock.assert_awaited_once_with(optimizer_settings.optimizer_ids)


async def test_resume_normal_mode(optimizer_inputs):
    tentacles_setup_config, trading_mode = optimizer_inputs
    optimizer_settings = bot_module_api.create_strategy_optimizer_settings({
        enums.OptimizerConfig.OPTIMIZER_CONFIG.value: MOCKED_OPTIMIZER_CONFIG,
        enums.OptimizerConfig.DATA_FILES.value: ["data_file"],
        enums.OptimizerConfig.OPTIMIZER_IDS.value: [1],
        enums.OptimizerConfig.RANDOMLY_CHOSE_RUNS.value: False,
        enums.OptimizerConfig.START_TIMESTAMP.value: 1661990400,
        enums.OptimizerConfig.END_TIMESTAMP.value: 1667347200,
        enums.OptimizerConfig.IDLE_CORES.value: 5,
        enums.OptimizerConfig.NOTIFY_WHEN_COMPLETE.value: True,
        enums.OptimizerConfig.MODE.value: enums.OptimizerModes.NORMAL.value,
        enums.OptimizerConfig.EMPTY_THE_QUEUE.value: False,
    })
    optimizer = bot_module_api.create_design_strategy_optimizer(
        trading_mode,
        optimizer_settings,
        None,
        tentacles_setup_config,
    )
    with mock.patch.object(optimizer, "_get_total_nb_runs",
                           mock.AsyncMock(return_value=5)) as _get_total_nb_runs_mock, \
            mock.patch.object(optimizer, "multi_processed_optimize",
                              mock.AsyncMock(return_value=True)) as multi_processed_optimize_mock:
        assert await optimizer.resume(optimizer_settings) is True
        _get_total_nb_runs_mock.assert_called_once_with(optimizer_settings.optimizer_ids)
        multi_processed_optimize_mock.assert_awaited_once_with(optimizer_settings)




#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
from setuptools import find_packages
from setuptools import setup
from octobot import PROJECT_NAME, AUTHOR, VERSION

PACKAGES = find_packages(exclude=["tentacles*", "tests", ])

# long description from README file
with open('README.md', encoding='utf-8') as f:
    DESCRIPTION = f.read()


def ignore_git_requirements(requirements):
    return [requirement for requirement in requirements if "git+" not in requirement]


REQUIRED = ignore_git_requirements(open('requirements.txt').readlines())
REQUIRES_PYTHON = '>=3.10'

setup(
    name=PROJECT_NAME,
    version=VERSION,
    url='https://github.com/Drakkar-Software/OctoBot',
    license='GPL-3.0',
    author=AUTHOR,
    author_email='contact@drakkar.software',
    description='Cryptocurrencies alert / trading bot',
    py_modules=['start'],
    packages=PACKAGES,
    package_data={
        "": ["config/*", "strategy_optimizer/optimizer_data_files/*"],
    },
    long_description=DESCRIPTION,
    long_description_content_type='text/markdown',
    tests_require=["pytest"],
    test_suite="tests",
    zip_safe=False,
    install_requires=REQUIRED,
    python_requires=REQUIRES_PYTHON,
    entry_points={
        'console_scripts': [
            PROJECT_NAME + ' = octobot.cli:main'
        ]
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3.10',
    ],
)



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import copy

from tests.unit_tests.community.errors_upload import basic_error, exception_error, ERROR_TITLE, ERROR_METRICS_ID, \
    ERROR_TIME


def test_constructor_with_basic_error(basic_error):
    assert basic_error.error is None
    assert basic_error.title is ERROR_TITLE
    assert basic_error.first_timestamp == ERROR_TIME
    assert basic_error.last_timestamp == ERROR_TIME
    assert basic_error.count == 1
    assert basic_error.metrics_id is ERROR_METRICS_ID
    assert basic_error.type == ""
    assert basic_error.stacktrace == []


def test_constructor_with_exception_error(exception_error):
    assert isinstance(exception_error.error, ZeroDivisionError)
    assert exception_error.title is ERROR_TITLE
    assert exception_error.first_timestamp == ERROR_TIME
    assert exception_error.last_timestamp == ERROR_TIME
    assert exception_error.count == 1
    assert exception_error.metrics_id is ERROR_METRICS_ID
    assert exception_error.type == ZeroDivisionError.__name__
    assert len(exception_error.stacktrace) == 5


def test_to_dict(exception_error):
    dict_repr = exception_error.to_dict()
    assert len(dict_repr) == 7
    assert all(bool(v) for v in dict_repr.values())


def test_is_equivalent(basic_error, exception_error):
    assert not basic_error.is_equivalent(exception_error)
    assert not exception_error.is_equivalent(basic_error)
    assert exception_error.is_equivalent(exception_error)
    assert basic_error.is_equivalent(basic_error)

    exception_error2 = copy.deepcopy(exception_error)
    basic_error2 = copy.deepcopy(basic_error)

    exception_error2.count += 1
    assert exception_error.is_equivalent(exception_error2)
    exception_error2.title += "1"
    assert not exception_error.is_equivalent(exception_error2)

    basic_error2.type = NotImplementedError
    assert not basic_error.is_equivalent(basic_error2)


def test_merge_equivalent(basic_error, exception_error):
    exception_error.last_timestamp = ERROR_TIME + 1
    basic_error.merge_equivalent(exception_error)
    assert basic_error.first_timestamp == ERROR_TIME
    assert basic_error.last_timestamp == ERROR_TIME + 1
    assert basic_error.count == 2
    exception_error.merge_equivalent(basic_error)
    assert exception_error.first_timestamp == ERROR_TIME
    assert exception_error.last_timestamp == ERROR_TIME + 1
    assert exception_error.count == 3



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import math

import octobot_commons.constants as commons_constants
import octobot_trading.api as trading_api
import octobot_trading.enums as trading_enum
import octobot_trading.constants as trading_constants
import octobot_trading.personal_data as trading_personal_data


def check_order_limits(order, market_status):
    symbol_market_limits = market_status[trading_enum.ExchangeConstantsMarketStatusColumns.LIMITS.value]
    limit_amount = symbol_market_limits[trading_enum.ExchangeConstantsMarketStatusColumns.LIMITS_AMOUNT.value]
    limit_cost = symbol_market_limits[trading_enum.ExchangeConstantsMarketStatusColumns.LIMITS_COST.value]
    limit_price = symbol_market_limits[trading_enum.ExchangeConstantsMarketStatusColumns.LIMITS_PRICE.value]

    min_quantity = limit_amount[trading_enum.ExchangeConstantsMarketStatusColumns.LIMITS_AMOUNT_MIN.value]
    max_quantity = limit_amount[trading_enum.ExchangeConstantsMarketStatusColumns.LIMITS_AMOUNT_MAX.value]
    min_cost = limit_cost[trading_enum.ExchangeConstantsMarketStatusColumns.LIMITS_COST_MIN.value]
    max_cost = limit_cost[trading_enum.ExchangeConstantsMarketStatusColumns.LIMITS_COST_MAX.value]
    min_price = limit_price[trading_enum.ExchangeConstantsMarketStatusColumns.LIMITS_PRICE_MIN.value]
    max_price = limit_price[trading_enum.ExchangeConstantsMarketStatusColumns.LIMITS_PRICE_MAX.value]
    maximal_price_digits = market_status[trading_enum.ExchangeConstantsMarketStatusColumns.PRECISION.value][
        trading_enum.ExchangeConstantsMarketStatusColumns.PRECISION_PRICE.value]
    maximal_volume_digits = market_status[trading_enum.ExchangeConstantsMarketStatusColumns.PRECISION.value][
        trading_enum.ExchangeConstantsMarketStatusColumns.PRECISION_AMOUNT.value]
    order_cost = order.origin_price * order.origin_quantity

    assert order_cost <= max_cost
    assert order_cost >= min_cost
    assert order.origin_price <= max_price
    assert order.origin_price >= min_price
    assert str(order.origin_price)[::-1].find(".") <= maximal_price_digits
    assert order.origin_quantity <= max_quantity
    assert order.origin_quantity >= min_quantity
    assert str(order.origin_quantity)[::-1].find(".") <= maximal_volume_digits


def check_oco_order_group(order, order_type, order_price, market_status):
    orders_from_group = [found_order
                         for found_order in order.order_group.get_group_open_orders()
                         if found_order is not order]
    assert len(orders_from_group) == 1
    other_order_from_oco_group = orders_from_group[0]
    assert other_order_from_oco_group.exchange_manager == order.exchange_manager
    assert other_order_from_oco_group.trader == order.trader
    assert other_order_from_oco_group.order_type == order_type
    assert other_order_from_oco_group.created_last_price == order.created_last_price
    assert other_order_from_oco_group.origin_price == order_price
    assert other_order_from_oco_group.order_group is order.order_group
    assert other_order_from_oco_group.created_last_price == order.created_last_price
    assert other_order_from_oco_group.currency == order.currency
    assert other_order_from_oco_group.fee == order.fee
    assert other_order_from_oco_group.filled_price == order.filled_price
    assert other_order_from_oco_group.filled_quantity == order.filled_quantity
    assert other_order_from_oco_group.status == order.status
    assert other_order_from_oco_group.symbol == order.symbol
    check_order_limits(order, market_status)


def check_orders(orders, evaluation, state, nb_orders, market_status):
    if state == trading_enum.EvaluatorStates.NEUTRAL.value or state is None:
        assert orders == []
    else:
        if math.isnan(evaluation):
            assert orders == []
        elif state is None or isinstance(state, (int, float, dict)) or state not in [s.value for s in
                                                                                     trading_enum.EvaluatorStates]:
            assert orders == []
        else:
            assert (not orders and nb_orders == 0) or (len(orders) == nb_orders) \
                   or ((len(orders) == 0 or len(orders) == 1) and nb_orders == "unknown")
            if orders:
                order = orders[0]
                if order.order_type in (
                trading_enum.TraderOrderType.SELL_MARKET, trading_enum.TraderOrderType.BUY_MARKET):
                    assert order.status == trading_enum.OrderStatus.FILLED
                    assert order.simulated is True
                    assert order.fee
                    assert order.filled_price > trading_constants.ZERO
                    assert order.filled_quantity == order.origin_quantity
                else:
                    assert order.status == trading_enum.OrderStatus.OPEN
                    assert order.simulated is True
                    assert order.fee is None
                    assert order.filled_price == trading_constants.ZERO
                    assert order.filled_quantity == order.origin_quantity

                if state == trading_enum.EvaluatorStates.VERY_SHORT.value:
                    assert isinstance(order, trading_personal_data.SellMarketOrder)
                    assert order.side == trading_enum.TradeOrderSide.SELL
                    assert order.order_type == trading_enum.TraderOrderType.SELL_MARKET
                elif state == trading_enum.EvaluatorStates.SHORT.value:
                    assert isinstance(order, trading_personal_data.SellLimitOrder)
                    assert order.side == trading_enum.TradeOrderSide.SELL
                    assert order.order_type == trading_enum.TraderOrderType.SELL_LIMIT
                elif state == trading_enum.EvaluatorStates.VERY_LONG.value:
                    assert isinstance(order, trading_personal_data.BuyMarketOrder)
                    assert order.side == trading_enum.TradeOrderSide.BUY
                    assert order.order_type == trading_enum.TraderOrderType.BUY_MARKET
                elif state == trading_enum.EvaluatorStates.LONG.value:
                    assert isinstance(order, trading_personal_data.BuyLimitOrder)
                    assert order.side == trading_enum.TradeOrderSide.BUY
                    assert order.order_type == trading_enum.TraderOrderType.BUY_LIMIT

                check_order_limits(order, market_status)


def check_portfolio(portfolio, initial_portfolio, orders, only_positivity=False):
    if orders:
        orders_market_amount = trading_constants.ZERO
        orders_currency_amount = trading_constants.ZERO
        market = orders[0].market
        order_symbol = orders[0].currency
        for order in orders:
            assert order.market == market
            assert order.currency == order_symbol
            if order.side == trading_enum.TradeOrderSide.BUY:
                orders_market_amount += order.origin_quantity * order.origin_price
            else:
                orders_currency_amount += order.origin_quantity
            for symbol in [market, order_symbol]:
                assert portfolio.get_currency_portfolio(symbol).total >= trading_constants.ZERO
                assert portfolio.get_currency_portfolio(symbol).available >= trading_constants.ZERO
                if not only_positivity:
                    if order.order_type in (
                    trading_enum.TraderOrderType.SELL_MARKET, trading_enum.TraderOrderType.BUY_MARKET) and \
                            symbol in [order.market, order.currency]:
                        # order is filled
                        assert initial_portfolio.get_currency_portfolio(symbol) != \
                               portfolio.get_currency_portfolio(symbol)
                    else:
                        if order_symbol == symbol:
                            assert initial_portfolio.get_currency_portfolio(symbol).total == \
                                   portfolio.get_currency_portfolio(symbol).total
                            assert initial_portfolio.get_currency_portfolio(symbol).available - orders_currency_amount \
                                   == portfolio.get_currency_portfolio(symbol).available
                        elif market == symbol:
                            assert initial_portfolio.get_currency_portfolio(market).total == \
                                   portfolio.get_currency_portfolio(market).total
                            assert initial_portfolio.get_currency_portfolio(market).available - orders_market_amount \
                                   == portfolio.get_currency_portfolio(market).available


async def fill_orders(orders, trader):
    if orders:
        assert trading_api.get_open_orders(trader.exchange_manager)
        for order in orders:
            order.filled_price = order.origin_price
            order.filled_quantity = order.origin_quantity
            await order.on_fill(force_fill=True)
            check_portfolio(trader.exchange_manager.exchange_personal_data.portfolio_manager.portfolio,
                            None, orders, True)
        assert len(trading_api.get_open_orders(trader.exchange_manager)) == 0



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import hashlib
import json
import multiprocessing
import os
import itertools
import enum
import queue
import copy
import asyncio
import concurrent.futures
import shutil
import time
import numpy
import logging
import ctypes
import random

import octobot.strategy_optimizer.optimizer_settings as optimizer_settings_import
import octobot.strategy_optimizer.optimizer_filter as optimizer_filter
import octobot.enums as enums
import octobot_commons.optimization_campaign as optimization_campaign
import octobot_commons.constants as commons_constants
import octobot_commons.enums as commons_enums
import octobot_commons.errors as commons_errors
import octobot_commons.logging as commons_logging
import octobot_commons.multiprocessing_util as multiprocessing_util
import octobot_commons.databases as databases
import octobot_commons.dict_util as dict_util
import octobot_backtesting.errors as backtesting_errors
import octobot_tentacles_manager.api as tentacles_manager_api
import octobot_tentacles_manager.constants as tentacles_manager_constants
import octobot_services.api as services_api
import octobot_services.enums as services_enums


class ConfigTypes(enum.Enum):
    NUMBER = "number"
    BOOLEAN = "boolean"
    OPTIONS = "options"
    UNKNOWN = "unknown"


class NoMoreRunError(Exception):
    pass


class StrategyDesignOptimizer:
    ALLOWED_IN_FACTORY = True

    SHARED_KEEP_RUNNING_KEY = "keep_running"
    SHARED_RUN_TIMES_KEY = "run_times"
    SHARED_RUNS_QUEUES_KEY = "runs_queues"
    START_QUEUE_KEY = "start_queue"
    DONE_QUEUE_KEY = "done_queue"

    RUN_SCHEDULE_TABLE = "schedule"
    CONFIG_KEY = "key"
    CONFIG_USER_INPUT = "user_input"
    CONFIG_USER_INPUTS = "user_inputs"
    CONFIG_FILTER_SETTINGS = "filters_settings"
    CONFIG_VALUE = "value"
    CONFIG_TENTACLE = "tentacle"
    CONFIG_NESTED_TENTACLE_SEPARATOR = "_------_"
    CONFIG_TYPE = "type"
    CONFIG_ENABLED = "enabled"
    CONFIG_MIN = "min"
    CONFIG_MAX = "max"
    CONFIG_STEP = "step"
    CONFIG_RUNS = "runs"
    CONFIG_ID = "id"
    CONFIG_DELETED = "deleted"
    CONFIG_DELETE_EVERY_RUN = "delete_every_run"
    CONFIG_ROLE = "role"
    LAST_CREATED_QUEUE = f"last_created_queue{commons_constants.CONFIG_FILE_EXT}"
    LAST_CREATED_QUEUE_CONFIG = f"last_created_queue_config{commons_constants.CONFIG_FILE_EXT}"

    def __init__(self, trading_mode, config, tentacles_setup_config, optimizer_settings=None):
        self.logger = commons_logging.get_logger(self.__class__.__name__)
        self.config = config
        self.base_tentacles_setup_config = tentacles_setup_config
        self.trading_mode = trading_mode
        self.optimizer_settings: optimizer_settings_import.OptimizerSettings = \
            optimizer_settings or optimizer_settings_import.OptimizerSettings()
        self.current_backtesting_id = None
        self.runs_schedule = None
        self.optimization_campaign_name = optimization_campaign.OptimizationCampaign.get_campaign_name(
            tentacles_setup_config
        )
        self.run_dbs_identifier = databases.RunDatabasesIdentifier(self.trading_mode, self.optimization_campaign_name)

        self.is_computing = False
        self.is_finished = False
        self.total_nb_runs = 0
        self.active_processes_count = 0
        self.keep_running = True
        self.process_pool_handle = None
        self.average_run_time = 0
        self.first_iteration_time = 0

    async def initialize(self, is_resuming):
        if not is_resuming:
            await self.generate_and_save_run()

    def get_name(self):
        return f"{self.trading_mode.get_name()}_{self.__class__.__name__}"

    async def _create_run_queues(self, optimizer_id, run_data):
        run_data_by_hash = self._get_optimizer_runs_details_and_hashes(run_data) if run_data \
            else await self._read_optimizer_runs_details_and_hashes(optimizer_id)
        start_queue = multiprocessing.Queue(len(run_data_by_hash))
        done_queue = multiprocessing.Queue(len(run_data_by_hash))
        for run_hash in run_data_by_hash:
            start_queue.put(run_hash)
        return {
            self.START_QUEUE_KEY: start_queue,
            self.DONE_QUEUE_KEY: done_queue
        }

    async def multi_processed_optimize(self, optimizer_settings, run_data_by_optimizer_id=None):
        success = True
        optimizer_ids = optimizer_settings.optimizer_ids or [optimizer_settings.optimizer_id]
        run_data_by_optimizer_id = run_data_by_optimizer_id or {}
        self.is_computing = True
        self.is_finished = False
        self.average_run_time = 0
        self.active_processes_count = multiprocessing.cpu_count() - abs(optimizer_settings.required_idle_cores)
        self.total_nb_runs = await self._get_total_nb_runs(optimizer_ids)
        global_t0 = time.time()
        lock = multiprocessing.RLock()
        shared_keep_running = multiprocessing.Value(ctypes.c_bool, True)
        shared_run_time = multiprocessing.Array(ctypes.c_float, [0.0 for _ in range(self.active_processes_count)])
        try:
            async for selected_optimizer_ids in self._all_optimizer_ids(optimizer_ids,
                                                                        optimizer_settings.empty_the_queue):
                run_queues_by_optimizer_id = {
                    optimizer_id: await self._create_run_queues(optimizer_id,
                                                                run_data_by_optimizer_id.get(optimizer_id))
                    for optimizer_id in selected_optimizer_ids
                }
                try:
                    await self._run_multi_processed_optimizer(
                        optimizer_settings, lock,
                        shared_keep_running, shared_run_time,
                        run_queues_by_optimizer_id
                    )
                finally:
                    # properly empty and close queues to avoid underlying thread issues
                    for optimizer_id, run_queues in run_queues_by_optimizer_id.items():
                        try:
                            await self._update_runs_from_done_queue(run_queues, optimizer_id)
                        except Exception as e:
                            self.logger.exception(e, True, f"Error on final db queue update: {e}")
                        for run_queue in run_queues.values():
                            # empty queues
                            while not run_queue.empty():
                                run_queue.get()
                            run_queue.close()
                            run_queue.cancel_join_thread()
        except Exception as e:
            self.logger.exception(e, True, f"Error when running optimizer processes: {e}")
            success = False
        finally:
            if optimizer_settings.notify_when_complete:
                await self._send_optimizer_finished_notification()
            self.process_pool_handle = None
            self.is_computing = False
            self.is_finished = True
        self.logger.info(f"Optimizer runs complete in {time.time() - global_t0} seconds.")
        return success

    async def _run_multi_processed_optimizer(self, optimizer_settings,
                                             lock, shared_keep_running, shared_run_time,
                                             run_queues_by_optimizer_id):
        with multiprocessing_util.registered_lock_and_shared_elements(
                commons_enums.MultiprocessingLocks.DBLock.value,
                lock,
                {
                    self.SHARED_KEEP_RUNNING_KEY: shared_keep_running,
                    self.SHARED_RUN_TIMES_KEY: shared_run_time,
                    self.SHARED_RUNS_QUEUES_KEY: run_queues_by_optimizer_id,
                }), \
                concurrent.futures.ProcessPoolExecutor(
                    max_workers=self.active_processes_count,
                    initializer=multiprocessing_util.register_lock_and_shared_elements,
                    initargs=(commons_enums.MultiprocessingLocks.DBLock.value,
                              lock,
                              {
                                  self.SHARED_KEEP_RUNNING_KEY: shared_keep_running,
                                  self.SHARED_RUN_TIMES_KEY: shared_run_time,
                                  self.SHARED_RUNS_QUEUES_KEY: run_queues_by_optimizer_id,
                              })) as pool:
            coros = []
            self.logger.info(f"Dispatching optimizer backtesting runs into {self.active_processes_count} "
                             f"parallel processes (based on the current computer physical processors).")
            if self.active_processes_count < 1:
                idle_cores_message = f"Requiring to leave {optimizer_settings.required_idle_cores} idle core. " \
                    if optimizer_settings.required_idle_cores else ""
                raise RuntimeError(f"{idle_cores_message}At lease one core is required to "
                                   f"start a strategy optimizer. "
                                   f"There are {multiprocessing.cpu_count()} total available cores on this computer.")
            for index in range(self.active_processes_count):
                coros.append(
                    asyncio.get_event_loop().run_in_executor(
                        pool,
                        self.find_optimal_configuration_wrapper,
                        optimizer_settings.data_files,
                        index == 0,
                        optimizer_settings.start_timestamp,
                        optimizer_settings.end_timestamp
                    )
                )
            self.process_pool_handle = await asyncio.gather(*coros)

    async def _all_optimizer_ids(self, prioritized_ids, empty_the_queue):
        if prioritized_ids:
            yield prioritized_ids
        if empty_the_queue:
            updated_ids = await self.get_queued_optimizer_ids()
            while self._should_keep_running() and updated_ids:
                # reselect from database in case new runs were added
                yield updated_ids
                updated_ids = await self.get_queued_optimizer_ids()

    def _init_optimizer_process_logger(self):
        # init basic logger
        logging.basicConfig(
            format=f'[{self.__class__.__name__} %(process)d] %(levelname)-6s %(name)-20s '
                   f'%(filename)-s:%(lineno)-8s %(message)s',
            level=logging.ERROR
        )

    def find_optimal_configuration_wrapper(self,
                                           data_files,
                                           update_database=False,
                                           start_timestamp=None,
                                           end_timestamp=None):
        self._init_optimizer_process_logger()
        run_queues_by_optimizer_id = multiprocessing_util.get_shared_element(self.SHARED_RUNS_QUEUES_KEY)
        asyncio.run(self.find_optimal_configuration(run_queues_by_optimizer_id,
                                                    data_files,
                                                    update_database=update_database,
                                                    start_timestamp=start_timestamp,
                                                    end_timestamp=end_timestamp))

    async def find_optimal_configuration(self,
                                         run_queues_by_optimizer_id,
                                         data_files,
                                         update_database=False,
                                         start_timestamp=None,
                                         end_timestamp=None):
        # need to load tentacles when in new process
        tentacles_manager_api.reload_tentacle_info()
        run_queues = optimizer_id = None
        try:
            for optimizer_id, run_queues in run_queues_by_optimizer_id.items():
                await self._execute_optimizer_run(optimizer_id, run_queues, update_database, data_files,
                                                  start_timestamp, end_timestamp)

        except concurrent.futures.CancelledError:
            self.logger.info("Cancelled run")
            if update_database and run_queues is not None and optimizer_id is not None:
                # stopped run: update queue
                await self._update_runs_from_done_queue(run_queues, optimizer_id)
            raise

    async def _read_optimizer_runs_details_and_hashes(self, optimizer_id):
        async with databases.DBReader.database(self.run_dbs_identifier.get_optimizer_runs_schedule_identifier()) \
                as reader:
            run_data = await self._get_run_data(optimizer_id, reader)
        try:
            return self._get_optimizer_runs_details_and_hashes(run_data[0][self.CONFIG_RUNS])
        except IndexError:
            raise NoMoreRunError

    def _get_optimizer_runs_details_and_hashes(self, run_data):
        if run_data:
            return {
                self.get_run_hash(run_details): run_details
                for run_details in run_data.values()
            }
        raise NoMoreRunError

    async def _execute_optimizer_run(self, optimizer_id, run_queues, update_database, data_files,
                                     start_timestamp, end_timestamp):
        if update_database and optimizer_id is None:
            # should not happen but in case it does, drop the run
            await self.drop_optimizer_run_from_queue(optimizer_id)
            return
        try:
            run_data_by_hash = await self._read_optimizer_runs_details_and_hashes(optimizer_id)
            t_start = time.time()
            last_db_update_time = t_start
            while self._should_keep_running():
                await self.run_single_iteration(optimizer_id,
                                                run_queues,
                                                run_data_by_hash,
                                                data_files,
                                                start_timestamp=start_timestamp,
                                                end_timestamp=end_timestamp)
                if self.first_iteration_time == 0:
                    self.first_iteration_time = time.time() - t_start
                    self._register_run_time(self.first_iteration_time)
                if update_database and time.time() - last_db_update_time > self.optimizer_settings.db_update_period:
                    await self._update_runs_from_done_queue(run_queues, optimizer_id)
                    last_db_update_time = time.time()
            if update_database:
                # stopped run: update queue
                await self._update_runs_from_done_queue(run_queues, optimizer_id)
        except NoMoreRunError:
            await self.drop_optimizer_run_from_queue(optimizer_id)
        except Exception as e:
            self.logger.exception(e, True, f"Error while running iteration: {e}")

    async def _update_runs_from_done_queue(self, run_queues, optimizer_id):
        completed_run_hashes = set()
        while not run_queues[self.DONE_QUEUE_KEY].empty():
            completed_run_hashes.add(run_queues[self.DONE_QUEUE_KEY].get(timeout=0.1))
        if not completed_run_hashes:
            return
        async with databases.DBWriterReader.database(
                self.run_dbs_identifier.get_optimizer_runs_schedule_identifier(), with_lock=True) \
                as writer_reader:
            run_data = await self._get_run_data_from_db(optimizer_id, writer_reader)
            if not run_data:
                # db already empty, nothing to do
                return
            updated_queue = run_data[0]
            runs = [run_details
                    for run_details in updated_queue[self.CONFIG_RUNS].values()
                    if self.get_run_hash(run_details) not in completed_run_hashes]
            updated_queue[self.CONFIG_RUNS] = {
                index: run
                for index, run in enumerate(runs)
            }
            query = await writer_reader.search()
            if runs and runs[0]:
                await writer_reader.update(self.RUN_SCHEDULE_TABLE, updated_queue,
                                           query.id == updated_queue["id"])

    async def drop_optimizer_run_from_queue(self, optimizer_id):
        async with databases.DBWriter.database(
                self.run_dbs_identifier.get_optimizer_runs_schedule_identifier(), with_lock=True) as writer:
            query = await writer.search()
            await writer.delete(self.RUN_SCHEDULE_TABLE, query.id == optimizer_id)

    def _should_keep_running(self):
        try:
            return multiprocessing_util.get_shared_element(self.SHARED_KEEP_RUNNING_KEY).value
        except KeyError:
            return self.keep_running

    def _register_run_time(self, run_time):
        try:
            shared_run_times = multiprocessing_util.get_shared_element(self.SHARED_RUN_TIMES_KEY)
            with shared_run_times:
                for index, value in enumerate(shared_run_times):
                    if value == 0.0:
                        shared_run_times[index] = run_time
                        return
        except KeyError:
            pass

    def get_average_run_time(self):
        if self.average_run_time == 0:
            try:
                run_times = [v for v in multiprocessing_util.get_shared_element(self.SHARED_RUN_TIMES_KEY)]
            except KeyError:
                run_times = []
            set_run_times = [run_time for run_time in run_times if run_time > 0]
            if not set_run_times:
                return 0
            average_run_time = numpy.average(set_run_times)
            if len(set_run_times) >= self.active_processes_count:
                self.average_run_time = average_run_time
            return average_run_time
        return self.average_run_time

    def cancel(self):
        self.keep_running = False
        try:
            multiprocessing_util.get_shared_element(self.SHARED_KEEP_RUNNING_KEY).value = False
        except KeyError:
            pass
        if self.process_pool_handle is not None:
            self.process_pool_handle.cancel()

    async def _get_remaining_runs_count_from_multi_process_queue(self, optimizer_ids):
        try:
            if not isinstance(optimizer_ids, list):
                optimizer_ids = [optimizer_ids]
            run_queues_by_optimizer_id = multiprocessing_util.get_shared_element(self.SHARED_RUNS_QUEUES_KEY)
            return sum(run_queues[self.START_QUEUE_KEY].qsize()
                       for optimizer_id, run_queues in run_queues_by_optimizer_id.items()
                       if optimizer_id in optimizer_ids or self.optimizer_settings.empty_the_queue)
        except KeyError:
            return 0

    async def get_overall_progress(self):
        if self.optimizer_settings.optimizer_id is None:
            remaining_runs = await self._get_remaining_runs_count_from_multi_process_queue(
                self.optimizer_settings.optimizer_ids)
        else:
            remaining_runs = await self._get_remaining_runs_count_from_multi_process_queue(
                self.optimizer_settings.optimizer_id)
        if self.is_computing and remaining_runs == 0:
            remaining_runs = 1
        done_runs = self.total_nb_runs - remaining_runs
        remaining_time = remaining_runs * self.get_average_run_time() / self.active_processes_count \
            if self.active_processes_count else 0
        return int(done_runs / self.total_nb_runs * 100) if self.total_nb_runs else 0, remaining_time

    async def is_in_progress(self):
        return self.is_computing

    def get_current_test_suite_progress(self):
        return 0

    def get_errors_description(self):
        return ""

    async def _get_run_data(self, optimizer_id, reader):
        if self.runs_schedule:
            return [self.runs_schedule]
        return await self._get_run_data_from_db(optimizer_id, reader)

    async def _get_run_data_from_db(self, optimizer_id, reader):
        if optimizer_id is None:
            raise RuntimeError("No optimizer id")
        query = await reader.search()
        return await reader.select(self.RUN_SCHEDULE_TABLE, query.id == optimizer_id)

    async def get_queued_optimizer_ids(self):
        return [
            run.get(self.CONFIG_ID, None)
            for run in (await self.get_run_queue(self.trading_mode, self.optimization_campaign_name))
        ]

    def _get_optimization_func(self, optimizer_settings: optimizer_settings_import.OptimizerSettings):
        if optimizer_settings.optimizer_mode == enums.OptimizerModes.NORMAL.value:
            return self.multi_processed_optimize
        return None

    async def resume(self, optimizer_settings: optimizer_settings_import.OptimizerSettings):
        self.total_nb_runs = await self._get_total_nb_runs(optimizer_settings.optimizer_ids)
        if optimizer_func := self._get_optimization_func(optimizer_settings):
            return await optimizer_func(optimizer_settings)
        else:
            raise NotImplementedError(f"Unknown optimizer mode: {optimizer_settings.optimizer_mode}")

    def _get_from_run_results(self, run_data_elements, all_run_results):
        if all_run_results:
            yielded_count = 1
            for run_data_from_runs in reversed(all_run_results):
                for to_find_run_data in run_data_elements.values():
                    if self._is_using_these_user_inputs(
                            run_data_from_runs,
                            to_find_run_data
                    ):
                        yield to_find_run_data, run_data_from_runs
                        if yielded_count == len(run_data_elements):
                            return
                        yielded_count += 1

    def _is_using_these_user_inputs(self, full_run_data, to_find_run_data):
        full_run_data_user_inputs = full_run_data[commons_enums.BacktestingMetadata.USER_INPUTS.value]
        for user_input_data in to_find_run_data:
            try:
                # todo handle user input with object config
                if full_run_data_user_inputs[user_input_data[self.CONFIG_TENTACLE][0]] \
                        [user_input_data[self.CONFIG_USER_INPUT]] != user_input_data[self.CONFIG_VALUE]:
                    return False
            except KeyError:
                return False
        return True

    async def _send_optimizer_finished_notification(self):
        await services_api.send_notification(
            services_api.create_notification(f"Your strategy optimizer just finished.",
                                             sound=services_enums.NotificationSound.FINISHED_PROCESSING,
                                             category=services_enums.NotificationCategory.OTHER)
        )

    async def _get_total_nb_runs(self, optimizer_ids):
        full_queue = await self.get_run_queue(self.trading_mode)
        if self.optimizer_settings.empty_the_queue:
            return sum(len(run[self.CONFIG_RUNS]) for run in full_queue)
        return sum(len(run[self.CONFIG_RUNS]) for run in full_queue if run[self.CONFIG_ID] in optimizer_ids)

    async def generate_and_save_run(self):
        taken_ids = await self.get_queued_optimizer_ids()
        self.optimizer_settings.optimizer_id = await self.run_dbs_identifier.generate_new_optimizer_id(taken_ids) \
            if self.optimizer_settings.optimizer_id is None else self.optimizer_settings.optimizer_id
        self.run_dbs_identifier.optimizer_id = self.optimizer_settings.optimizer_id
        await self.run_dbs_identifier.initialize()
        return await self._generate_and_store_backtesting_runs_schedule()

    @classmethod
    async def get_run_queue(cls, trading_mode, campaign_name=None):
        campaign_name = campaign_name or optimization_campaign.OptimizationCampaign.get_campaign_name()
        run_dbs_identifier = databases.RunDatabasesIdentifier(trading_mode, campaign_name)
        try:
            async with databases.DBReader.database(run_dbs_identifier.get_optimizer_runs_schedule_identifier(),
                                                   with_lock=True) as reader:
                return await reader.all(cls.RUN_SCHEDULE_TABLE)
        except commons_errors.DatabaseNotFoundError:
            return []
        except json.JSONDecodeError:
            return []

    @classmethod
    def get_run_hash(cls, run_data):
        return hashlib.sha256(json.dumps([
            {
                cls.CONFIG_USER_INPUT: run_data_item[cls.CONFIG_USER_INPUT],
                cls.CONFIG_TENTACLE: run_data_item[cls.CONFIG_TENTACLE],
                cls.CONFIG_VALUE: float(run_data_item[cls.CONFIG_VALUE])
                if isinstance(run_data_item[cls.CONFIG_VALUE], int) else run_data_item[cls.CONFIG_VALUE]
            }
            for run_data_item in run_data
        ]).encode()).hexdigest()

    @classmethod
    async def update_run_queue(cls, trading_mode, updated_queue):
        run_dbs_identifier = databases.RunDatabasesIdentifier(
            trading_mode, optimization_campaign.OptimizationCampaign.get_campaign_name())
        async with databases.DBWriterReader.database(run_dbs_identifier.get_optimizer_runs_schedule_identifier(),
                                                     with_lock=True) as writer_reader:
            query = await writer_reader.search()
            try:
                runs = updated_queue[cls.CONFIG_RUNS]
                if updated_queue[cls.CONFIG_DELETE_EVERY_RUN]:
                    # delete every run, just delete the whole optimizer run
                    await writer_reader.delete(cls.RUN_SCHEDULE_TABLE, query.id == updated_queue["id"])
                    return updated_queue
                db_optimizer_run = []
                try:
                    db_optimizer_run = \
                        (await writer_reader.select(cls.RUN_SCHEDULE_TABLE, query.id == updated_queue[cls.CONFIG_ID]))[
                            0]
                except json.JSONDecodeError:
                    pass
                existing_runs = set(
                    cls.get_run_hash(run_data)
                    for run_data in db_optimizer_run.get(cls.CONFIG_RUNS).values()
                )
                to_remove_indexes = []
                # remove deleted runs and runs that are not in the queue anymore (already running or done)
                for run_index, run_data in enumerate(copy.copy(runs)):
                    if cls.get_run_hash(run_data) in existing_runs:
                        for run_input in run_data:
                            if run_input.get(cls.CONFIG_DELETED, False):
                                to_remove_indexes.append(run_index)
                                break
                            run_input.pop(cls.CONFIG_DELETED)
                    else:
                        runs.remove(run_data)
                for index in reversed(to_remove_indexes):
                    runs.pop(index)
                # ensure runs are a dict
                updated_queue[cls.CONFIG_RUNS] = {
                    index: run
                    for index, run in enumerate(runs)
                }
                # replace queue by updated one to keep order
                query = await writer_reader.search()
                if runs and runs[0]:
                    await writer_reader.update(cls.RUN_SCHEDULE_TABLE, updated_queue,
                                               query.id == updated_queue["id"])
                else:
                    await writer_reader.delete(cls.RUN_SCHEDULE_TABLE, query.id == updated_queue["id"])
            except IndexError as e:
                # optimizer run not in db anymore
                pass
            return updated_queue

    async def _synchronized_pick_run(self, run_queue, run_details_by_hash):
        try:
            # acquire MultiprocessingLocks.DBLock to wait for re-ordering if it happens during a run
            with multiprocessing_util.get_lock(commons_enums.MultiprocessingLocks.DBLock.value):
                # wait for a very short  time to allow queue sync between processes
                selected_run_hash = run_queue.get(timeout=1)
            run_details = run_details_by_hash.pop(selected_run_hash, None)
            self.logger.info(f"Selecting: {run_details}")
            return selected_run_hash, run_details
        except queue.Empty:
            raise NoMoreRunError("Nothing to run")
        except Exception as e:
            self.logger.exception(e)

    async def run_single_iteration(self,
                                   optimizer_id,
                                   run_queues,
                                   run_data_by_hash,
                                   data_files,
                                   start_timestamp=None,
                                   end_timestamp=None):
        start_run = False
        selected_run_hash, run_details = await self._synchronized_pick_run(run_queues[self.START_QUEUE_KEY],
                                                                           run_data_by_hash)
        if run_details:
            # start backtesting run ids at 1
            backtesting_run_id = 1
            # always ensure backtesting id is unique
            run_dbs_identifier = databases.RunDatabasesIdentifier(
                self.trading_mode, self.optimization_campaign_name,
                optimizer_id=optimizer_id, backtesting_id=backtesting_run_id
            )
            # acquire MultiprocessingLocks.DBLock to avoid conflicts during multi processing runs
            with multiprocessing_util.get_lock(commons_enums.MultiprocessingLocks.DBLock.value):
                backtesting_run_id = await run_dbs_identifier.generate_new_backtesting_id()
                run_dbs_identifier.backtesting_id = backtesting_run_id
                await run_dbs_identifier.initialize()
            start_run = True
        if start_run:
            try:
                run_result = await self._run_with_config(optimizer_id, data_files, backtesting_run_id, run_details,
                                                         start_timestamp=start_timestamp, end_timestamp=end_timestamp)
                return run_result
            finally:
                if selected_run_hash is not None:
                    run_queues[self.DONE_QUEUE_KEY].put(selected_run_hash)
        raise NoMoreRunError("Nothing to run")

    async def _run_with_config(self, optimizer_id, data_files, run_id, run_config,
                               start_timestamp=None, end_timestamp=None):
        self.logger.debug(f"Running optimizer with id {optimizer_id} "
                          f"on backtesting {run_id} with config {run_config}")
        self._update_config_for_optimizer(optimizer_id, run_id)
        tentacles_setup_config = self._get_custom_tentacles_setup_config(optimizer_id, run_id, run_config)
        independent_backtesting = None
        try:
            import octobot.api.backtesting as octobot_backtesting_api
            # reset possible remaining caches
            await databases.CacheManager().reset()
            config_to_use = copy.deepcopy(self.config)
            independent_backtesting = octobot_backtesting_api.create_independent_backtesting(
                config_to_use,
                tentacles_setup_config,
                data_files,
                start_timestamp=start_timestamp,
                end_timestamp=end_timestamp,
                enforce_total_databases_max_size_after_run=False,
            )
            await octobot_backtesting_api.initialize_and_run_independent_backtesting(independent_backtesting,
                                                                                     log_errors=False)
            await octobot_backtesting_api.join_independent_backtesting(independent_backtesting, timeout=None)
            return independent_backtesting
        except backtesting_errors.MissingTimeFrame:
            # ignore this exception: is due to missing of the only required time frame
            return independent_backtesting
        except Exception as e:
            self.logger.exception(e, True, str(e))
            return independent_backtesting
        finally:
            if independent_backtesting is not None:
                await independent_backtesting.stop()

    def _update_config_for_optimizer(self, optimizer_id, run_id):
        self.config[commons_constants.CONFIG_OPTIMIZER_ID] = optimizer_id
        self.config[commons_constants.CONFIG_BACKTESTING_ID] = run_id

    def _updated_nested_tentacle_config(self, nested_tentacles, user_input, config_value, local_tentacle_config):
        cleaned_tentacle_name = nested_tentacles[0].replace(" ", "_")
        if cleaned_tentacle_name not in local_tentacle_config:
            local_tentacle_config[cleaned_tentacle_name] = {}
        if len(nested_tentacles) == 1:
            local_tentacle_config[cleaned_tentacle_name][user_input.replace(" ", "_")] = config_value
        else:
            self._updated_nested_tentacle_config(nested_tentacles[1:], user_input, config_value,
                                                 local_tentacle_config[cleaned_tentacle_name])

    def _get_custom_tentacles_setup_config(self, optimizer_id, run_id, run_config):
        local_tentacles_setup_config = copy.deepcopy(self.base_tentacles_setup_config)
        run_dbs_identifier = databases.RunDatabasesIdentifier(
            self.trading_mode, self.optimization_campaign_name, optimizer_id=optimizer_id, backtesting_id=run_id)
        run_folder = run_dbs_identifier.get_backtesting_run_folder()
        tentacles_setup_config_path = os.path.join(run_folder, commons_constants.CONFIG_TENTACLES_FILE)
        tentacles_manager_api.set_tentacles_setup_configuration_path(local_tentacles_setup_config,
                                                                     tentacles_setup_config_path)
        # save updated config for each custom parameter tentacle
        # and import config of remaining tentacles: add missing tentacles into tentacles_updates with nothing to update
        tentacles_updates = {
            tentacle: {}
            for tentacle in tentacles_manager_api.get_activated_tentacles(self.base_tentacles_setup_config)
            if tentacles_manager_api.has_profile_local_configuration(self.base_tentacles_setup_config, tentacle)
        }
        for input_config in run_config:
            self._updated_nested_tentacle_config(input_config[self.CONFIG_TENTACLE],
                                                 input_config[self.CONFIG_USER_INPUT],
                                                 input_config[self.CONFIG_VALUE],
                                                 tentacles_updates)
        self._ensure_local_config_folders(run_folder)
        for tentacle, updated_values in tentacles_updates.items():
            tentacle_class = tentacles_manager_api.get_tentacle_class_from_string(tentacle)
            current_config = tentacles_manager_api.get_tentacle_config(self.base_tentacles_setup_config,
                                                                       tentacle_class)
            dict_util.nested_update_dict(current_config, updated_values)
            tentacles_manager_api.update_tentacle_config(local_tentacles_setup_config, tentacle_class, current_config)
        return local_tentacles_setup_config

    @staticmethod
    def _ensure_local_config_folders(run_folder):
        if not os.path.exists(run_folder):
            os.makedirs(run_folder)
        tentacles_specific_config_folder = os.path.join(run_folder,
                                                        tentacles_manager_constants.TENTACLES_SPECIFIC_CONFIG_FOLDER)
        if not os.path.exists(tentacles_specific_config_folder):
            os.mkdir(tentacles_specific_config_folder)

    async def _generate_and_store_backtesting_runs_schedule(self):
        runs = self._generate_runs()
        await self._save_run_schedule(runs)
        await self._create_run_schedule_and_config_snapshot()
        return runs

    @staticmethod
    def shuffle_and_select_runs(runs, select_size=None) -> dict:
        shuffled_runs = list(runs.values())
        random.shuffle(shuffled_runs)
        selected_runs = shuffled_runs if select_size is None else shuffled_runs[:select_size]
        return {i: run for i, run in enumerate(selected_runs)}

    def _generate_runs(self):
        iterations = [i for i in self._get_config_possible_iterations() if i]
        runs = {
            index: run
            for index, run in enumerate(itertools.product(*iterations))
            if self._is_run_allowed(run)
        }
        if runs:
            shuffled_runs = self.shuffle_and_select_runs(runs, select_size=self.optimizer_settings.queue_size)
            for run in shuffled_runs.values():
                for run_input in run:
                    # do not store self.CONFIG_KEY
                    run_input.pop(self.CONFIG_KEY, None)
            return shuffled_runs
        raise RuntimeError("No optimizer run to schedule with this configuration")

    def _is_run_allowed(self, run):
        for run_filter_config in self.optimizer_settings.optimizer_config[self.CONFIG_FILTER_SETTINGS]:
            if self._is_filtered(run, run_filter_config):
                return False
        return True

    def _is_filtered(self, run, run_filter_config):
        operator_filter = self._parse_filter_entry(run, run_filter_config)
        return operator_filter.is_filtered()

    def _parse_filter_entry(self, run, run_filter_config):
        user_input_left_operand_key = run_filter_config["user_input_left_operand"][self.CONFIG_VALUE]
        user_input_right_operand_key = run_filter_config["user_input_right_operand"][self.CONFIG_VALUE]

        left_operand = user_input_right_operand = None
        text_right_operand = run_filter_config["text_right_operand"][self.CONFIG_VALUE]
        operator = run_filter_config["operator"][self.CONFIG_VALUE]
        for user_input in run:
            if user_input[self.CONFIG_KEY] == user_input_left_operand_key:
                left_operand = user_input[self.CONFIG_VALUE]
            if user_input[self.CONFIG_KEY] == user_input_right_operand_key:
                user_input_right_operand = user_input[self.CONFIG_VALUE]
        right_operand = user_input_right_operand if text_right_operand in ("null", "") else text_right_operand
        return optimizer_filter.OptimizerFilter(user_input_left_operand_key, user_input_right_operand_key,
                                                left_operand, right_operand,
                                                operator)

    def _get_config_possible_iterations(self):
        return [
            self._generate_possible_values(config_element)
            for config_element in self._get_config_elements()
        ]

    def _generate_possible_values(self, config_element):
        if not config_element:
            return {}
        values = []
        try:
            if config_element[self.CONFIG_ENABLED]:
                if config_element[self.CONFIG_TYPE] is ConfigTypes.OPTIONS:
                    values = [[value] for value in config_element[self.CONFIG_VALUE][self.CONFIG_VALUE]]
                if config_element[self.CONFIG_TYPE] is ConfigTypes.BOOLEAN:
                    values = config_element[self.CONFIG_VALUE][self.CONFIG_VALUE]
                if config_element[self.CONFIG_TYPE] is ConfigTypes.NUMBER:
                    config = config_element[self.CONFIG_VALUE][self.CONFIG_VALUE]
                    values = [v
                              for v in self._get_all_possible_values(config[self.CONFIG_MIN],
                                                                     config[self.CONFIG_MAX],
                                                                     config[self.CONFIG_STEP])
                              if v <= config[self.CONFIG_MAX]]
            return [
                {
                    self.CONFIG_USER_INPUT: config_element[self.CONFIG_USER_INPUT],
                    self.CONFIG_TENTACLE: config_element[self.CONFIG_VALUE][self.CONFIG_TENTACLE]
                    .split(self.CONFIG_NESTED_TENTACLE_SEPARATOR),
                    self.CONFIG_VALUE: value,
                    self.CONFIG_KEY: config_element[self.CONFIG_KEY]
                }
                for value in values
            ]
        except ZeroDivisionError as e:
            raise ZeroDivisionError("Step value has to be greater than 0") from e

    def _get_all_possible_values(self, start, stop, step):
        # use custom decimal function for precision instead of numpy.arange
        d_start = decimal.Decimal(str(start))
        d_stop = decimal.Decimal(str(stop))
        d_step = decimal.Decimal(str(step))
        return_type = self.get_accurate_number_type(start, stop, step)
        current = d_start
        while current <= d_stop:
            yield return_type(current)
            current += d_step

    @staticmethod
    def get_accurate_number_type(*values):
        return int if all(isinstance(e, int) for e in values) else float

    def _get_config_elements(self):
        for key, val in self.optimizer_settings.optimizer_config[self.CONFIG_USER_INPUTS].items():
            yield {
                self.CONFIG_KEY: key,
                self.CONFIG_ENABLED: val[self.CONFIG_ENABLED],
                self.CONFIG_USER_INPUT: val[self.CONFIG_USER_INPUT],
                self.CONFIG_VALUE: val,
                self.CONFIG_TYPE: self._get_config_type(val)
            }

    def _get_config_type(self, value):
        config_values = value[self.CONFIG_VALUE]
        if isinstance(config_values, list):
            if not config_values:
                return ConfigTypes.OPTIONS
            if isinstance(config_values[0], bool):
                return ConfigTypes.BOOLEAN
            else:
                return ConfigTypes.OPTIONS
        if isinstance(config_values, dict):
            try:
                if all(isinstance(config_values[key], (float, int))
                       for key in [self.CONFIG_MIN, self.CONFIG_MAX, self.CONFIG_STEP]):
                    return ConfigTypes.NUMBER
            except KeyError:
                pass
        return ConfigTypes.UNKNOWN

    async def _create_run_schedule_and_config_snapshot(self):
        # gives the optimizer folder as no backtesting id is set
        optimizer_target_identifier = self.run_dbs_identifier.get_backtesting_run_folder()
        try:
            shutil.copy(
                self.run_dbs_identifier.get_optimizer_runs_schedule_identifier(),
                os.path.join(optimizer_target_identifier, self.LAST_CREATED_QUEUE)
            )
            with open(os.path.join(optimizer_target_identifier, self.LAST_CREATED_QUEUE_CONFIG), "w") as fw:
                json.dump(self.optimizer_settings.optimizer_config, fw)
        except FileNotFoundError as err:
            self.logger.debug(f"Skipped run schedule and config save: {err}")

    async def _save_run_schedule(self, runs):
        self.runs_schedule = {
            self.CONFIG_RUNS: runs,
            self.CONFIG_ID: self.optimizer_settings.optimizer_id
        }
        self.total_nb_runs = len(runs)
        async with databases.DBWriterReader.database(self.run_dbs_identifier.get_optimizer_runs_schedule_identifier(),
                                                     with_lock=True) as writer_reader:
            try:
                existing_runs = await self._get_run_data_from_db(self.optimizer_settings.optimizer_id, writer_reader)
                if existing_runs:
                    merged_runs = {
                        self.get_run_hash(run_data): run_data
                        for run_data in existing_runs[0][self.CONFIG_RUNS].values()
                        if run_data
                    }
                    merged_runs.update({
                        self.get_run_hash(run_data): run_data
                        for run_data in self.runs_schedule[self.CONFIG_RUNS].values()
                        if run_data
                    })
                    self.runs_schedule[self.CONFIG_RUNS] = {
                        f"{index}": details
                        for index, details in enumerate(merged_runs.values())
                    }
                    await writer_reader.delete(self.RUN_SCHEDULE_TABLE, (await writer_reader.search()).id ==
                                               self.optimizer_settings.optimizer_id)
                await writer_reader.log(self.RUN_SCHEDULE_TABLE, self.runs_schedule)
            except json.JSONDecodeError:
                self.logger.error(f"Invalid data in run schedule, clearing runs.")
                # error in database, reset it
                await writer_reader.hard_reset()
                await writer_reader.log(self.RUN_SCHEDULE_TABLE, self.runs_schedule)



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import asyncio
import pytest

from octobot_commons.logging.logging_util import get_logger
from octobot_trading.api.exchange import cancel_ccxt_throttle_task
from tentacles.Services.Interfaces.web_interface import WebInterface
from octobot_commons.tests.test_config import load_test_config
from octobot.commands import start_bot
from octobot.logger import init_logger
from octobot.octobot import OctoBot
import octobot.community as community

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


@pytest.mark.timeout(12)
async def test_run_bot():
    # avoid web interface in this test
    WebInterface.enabled = False
    community.IdentifiersProvider.use_production()
    bot = OctoBot(load_test_config(dict_only=False), ignore_config=True)
    bot.task_manager.init_async_loop()
    await start_bot(bot, init_logger())
    await asyncio.sleep(10)
    await stop_bot(bot)


async def stop_bot(bot):
    # force all logger enable to display any error
    stop_logger = get_logger("StopBotLogger")
    import logging
    for logger in logging.Logger.manager.loggerDict.values():
        logger.disabled = False

    stop_logger.info("Stopping tasks...")
    await bot.stop()

    if bot.task_manager.tools_task_group:
        bot.task_manager.tools_task_group.cancel()

    # close community session
    if bot.community_handler:
        await bot.community_handler.stop_task()

    cancel_ccxt_throttle_task()

    stop_logger.info("Tasks stopped.")



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import json
import enum
import realtime


import octobot_commons.logging as logging


class CHANNEL_STATES(enum.Enum):
    JOINED = "JOINED"
    JOINING = "JOINING"
    CLOSED = "CLOSED"
    ERRORED = "ERRORED"


class AuthenticatedSupabaseRealtimeChannel(realtime.Channel):
    def __init__(self, socket: realtime.Socket, topic: str, schema: str, table_name: str, params: dict = {}):
        """
        from realtime.Channel docs:
        `Channel` is an abstraction for a topic listener for an existing socket connection.
        Each Channel has its own topic and a list of event-callbacks that responds to messages.
        Should only be instantiated through `connection.Socket().set_channel(topic)`
        Topic-Channel has a 1-many relationship.


        AuthenticatedSupabaseRealtimeChannel tries to expose a similar logic and public API
        as https://github.com/supabase/realtime-js/blob/master/src/RealtimeChannel.ts
        Re-implements async methods as they are not working as is.
        """
        super().__init__(socket, topic, params=params)
        self.schema = schema
        self.table_name = table_name
        self.auth_payload = {}  # missing in realtime.Channel
        self.joined_once = False
        self.state: CHANNEL_STATES = CHANNEL_STATES.CLOSED
        self.logger = logging.get_logger(self.__class__.__name__)

    async def subscribe(self, callback=None, system_callback=None) -> None:
        """
        fix of async def realtime.Channel._join(self) -> None:
        adds self.join_payload handling (required for auth on join)

        Coroutine that attempts to join Phoenix Realtime server via a certain topic
        :return: None
        """
        if callback:
            self.socket.register_subscribe_callback(self.topic, callback)
        if system_callback:
            self.socket.register_system_callback(self.topic, system_callback)
        join_req = dict(topic=self.topic, event="phx_join", payload=self._get_join_payload(), ref=self._get_ref())
        try:
            self.state = CHANNEL_STATES.JOINING
            await self.socket.ws_connection.send(json.dumps(join_req))
            self.state = CHANNEL_STATES.JOINED
            self.joined_once = True
        except Exception as err:
            self.state = CHANNEL_STATES.ERRORED
            self.logger.exception(err, True, f"Error when joining realtime channel: {err}")
            return

    def _get_join_payload(self):
        # "postgres_changes": [
        #     {
        #         "event": "*" | "INSERT" | "UPDATE" | "DELETE",
        #         "schema": string,
        #         "table": string,
        #         "filter": string + '=' + "eq" | "neq" | "gt" | "gte" | "lt" | "lte" | "in" + '.' + string
        #     }
        # ]
        join_payload = {
            "postgres_changes": [
                {
                    "event": listener.event,
                    "schema": self.schema,
                    "table": self.table_name,
                }
            ]
            for listener in self.listeners
        }
        return {**join_payload, **self.auth_payload}

    def on(self, event, callback):
        """
        Fix payload handling
        """
        async def _cb(payload):
            enriched_payload = {
                "schema": payload["schema"],
                "table": payload["table"],
                "commit_timestamp": payload["commit_timestamp"],
                "event_type": payload["type"],
                "new": {},
                "old": {},
            }
            enriched_payload = {**enriched_payload, **_get_payload_records(payload)}
            await callback(enriched_payload)

        cl = realtime.CallbackListener(event=event, callback=_cb)
        self.listeners.append(cl)
        return self

    def update_auth_payload(self, update: dict):
        self.auth_payload.update(update)

    async def auth(self):
        full_payload = {
            "topic": self.topic,
            "event": "access_token",
            "payload": self.auth_payload,
            "ref": self._get_ref(),
            # "join_ref": 1, #todo https://github.com/supabase/realtime-js/blob/master/src/lib/push.ts#L79
        }
        await self.socket.ws_connection.send(json.dumps(full_payload))

    def is_joined(self):
        return self.state is CHANNEL_STATES.JOINED

    def is_joining(self):
        return self.state is CHANNEL_STATES.JOINING

    def _get_ref(self):
        # see https://github.com/supabase/realtime-js/blob/master/src/RealtimeClient.ts#L290
        return None  # implement if necessary


def _get_payload_records(payload):
    records: dict = {"new": {}, "old": {}}
    if payload["type"] in ["INSERT", "UPDATE"]:
        records["new"] = payload["record"]
        # skip type conversion: error with jsonb and not necessary
        # realtime.convert_change_data(payload["columns"], payload["record"], {"skip_types": realtime.jsonb})
    if payload["type"] in ["UPDATE", "DELETE"]:
        records["old"] = payload["record"]
        # skip type conversion: error with jsonb and not necessary
        # realtime.convert_change_data(payload["columns"], payload["old_record"], {"skip_types": realtime.jsonb})
    return records



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import asyncio
import datetime
import json
import time
import typing
import logging

import aiohttp
import gotrue.errors
import supabase.lib.client_options

import octobot_commons.authentication as authentication
import octobot_commons.logging as commons_logging
import octobot_commons.profiles as commons_profiles
import octobot_commons.enums as commons_enums
import octobot_commons.constants as commons_constants
import octobot.constants as constants
import octobot.community.errors as errors
import octobot.community.supabase_backend.enums as enums
import octobot.community.supabase_backend.supabase_client as supabase_client
import octobot.community.supabase_backend.configuration_storage as configuration_storage


_INTERNAL_LOGGERS = [
    "httpx", "httpx._client",
    "httpcore.http11", "httpcore.http2", "httpcore.proxy", "httpcore.socks", "httpcore.connection"
]
# disable httpx info logs as it logs every request
commons_logging.set_logging_level(_INTERNAL_LOGGERS, logging.WARNING)


class CommunitySupabaseClient(supabase_client.AuthenticatedAsyncSupabaseClient):
    """
    Octobot Community layer added to supabase_client.AuthenticatedSupabaseClient
    """
    def __init__(
        self,
        supabase_url: str,
        supabase_key: str,
        storage: configuration_storage.SyncConfigurationStorage,
        options: supabase.lib.client_options.ClientOptions = supabase.lib.client_options.ClientOptions(),
    ):
        options.storage = storage   # use configuration storage
        self.event_loop = None
        super().__init__(supabase_url, supabase_key, options=options)
        self.is_admin = False
        self.production_anon_client = None

    async def sign_in(self, email: str, password: str) -> None:
        try:
            self.event_loop = asyncio.get_event_loop()
            self.auth.sign_in_with_password({
                "email": email,
                "password": password,
            })
        except gotrue.errors.AuthApiError as err:
            if "email" in str(err).lower():
                raise errors.EmailValidationRequiredError(err) from err
            raise authentication.FailedAuthentication(err) from err

    async def sign_up(self, email: str, password: str) -> None:
        try:
            self.event_loop = asyncio.get_event_loop()
            self.auth.sign_up({
                "email": email,
                "password": password,
                "options": {
                    "data": {
                        "hasRegisteredFromSelfHosted": True
                    }
                }
            })
        except gotrue.errors.AuthApiError as err:
            raise authentication.AuthenticationError(err) from err

    def sign_out(self) -> None:
        self.auth.sign_out()

    def restore_session(self):
        self.event_loop = asyncio.get_event_loop()
        self.auth.initialize_from_storage()
        if not self.is_signed_in():
            raise authentication.FailedAuthentication()

    async def sign_in_with_otp_token(self, token):
        self.event_loop = asyncio.get_event_loop()
        # restore saved session in case otp token fails
        saved_session = self.auth._storage.get_item(self.auth._storage_key)
        try:
            url = f"{self.auth_url}/verify?token={token}&type=magiclink"
            async with aiohttp.ClientSession() as client:
                resp = await client.get(url, allow_redirects=False)
                self.auth.initialize_from_url(
                    resp.headers["Location"].replace("#access_token", "?access_token").replace("#error", "?error")
                )
        except gotrue.errors.AuthImplicitGrantRedirectError as err:
            if saved_session:
                self.auth._storage.set_item(self.auth._storage_key, saved_session)
            raise authentication.AuthenticationError(err) from err

    def is_signed_in(self) -> bool:
        try:
            return self.auth.get_session() is not None
        except gotrue.errors.AuthApiError as err:
            commons_logging.get_logger(self.__class__.__name__).info(f"Authentication error: {err}")
            # remove invalid session from
            self.remove_session_details()
            return False

    def has_login_info(self) -> bool:
        return bool(self.auth._storage.get_item(self.auth._storage_key))

    async def update_metadata(self, metadata_update) -> dict:
        return self.auth.update_user({
            "data": metadata_update
        }).user.dict()

    async def get_user(self) -> dict:
        try:
            return (await self._get_user()).dict()
        except gotrue.errors.AuthApiError as err:
            if "missing" in str(err):
                raise errors.EmailValidationRequiredError(err) from err
            raise authentication.AuthenticationError(err) from err

    def sync_get_user(self) -> dict:
        return self.auth.get_user().user.dict()

    async def fetch_bot(self, bot_id) -> dict:
        try:
            # https://postgrest.org/en/stable/references/api/resource_embedding.html#hint-disambiguation
            return (await self.table("bots").select("*,bot_deployment:bot_deployments!bots_current_deployment_id_fkey(*)").eq(
                enums.BotKeys.ID.value, bot_id
            ).execute()).data[0]
        except IndexError:
            raise errors.BotNotFoundError(f"Can't find bot with id: {bot_id}")

    async def fetch_bots(self) -> list:
        return (await self.table("bots").select("*,bot_deployment:bot_deployments!bots_current_deployment_id_fkey(*)").execute()).data

    async def create_bot(self, deployment_type: enums.DeploymentTypes) -> dict:
        created_bot = (await self.table("bots").insert({
            enums.BotKeys.USER_ID.value: (await self._get_user()).id
        }).execute()).data[0]
        bot_id = created_bot[enums.BotKeys.ID.value]
        created_deployment = await self._create_deployment(
            deployment_type, bot_id, constants.VERSION
        )
        await self.table("bots").update({
            enums.BotKeys.CURRENT_DEPLOYMENT_ID.value: created_deployment[enums.BotDeploymentKeys.ID.value],
        }).eq(
            enums.BotDeploymentKeys.ID.value, bot_id
        ).execute()
        # fetch bot to fetch embed elements (like deployments)
        return await self.fetch_bot(bot_id)

    async def _create_deployment(self, deployment_type, bot_id, version):
        return (await self.table("bot_deployments").insert({
            enums.BotDeploymentKeys.TYPE.value: deployment_type.value,
            enums.BotDeploymentKeys.VERSION.value: version,
            enums.BotDeploymentKeys.BOT_ID.value: bot_id,
        }).execute()).data[0]

    async def update_bot(self, bot_id, bot_update) -> dict:
        await self.table("bots").update(bot_update).eq(enums.BotKeys.ID.value, bot_id).execute()
        # fetch bot to fetch embed elements (like deployments)
        return await self.fetch_bot(bot_id)

    async def update_deployment(self, deployment_id, deployment_update) -> dict:
        return (await self.table("bot_deployments").update(deployment_update).eq(
            enums.BotDeploymentKeys.ID.value, deployment_id
        ).execute()).data[0]

    async def delete_bot(self, bot_id) -> list:
        return (await self.table("bots").delete().eq(enums.BotKeys.ID.value, bot_id).execute()).data

    async def fetch_deployment_url(self, deployment_url_id) -> dict:
        try:
            return (await self.table("bot_deployment_urls").select("*").eq(
                enums.BotDeploymentURLKeys.ID.value, deployment_url_id
            ).execute()).data[0]
        except IndexError:
            raise errors.BotDeploymentURLNotFoundError(deployment_url_id)

    async def fetch_startup_info(self, bot_id) -> dict:
        return json.loads(
            (await self.postgres_functions().invoke("get_startup_info", {"body": {"bot_id": bot_id}}))["data"]
        )[0]

    async def fetch_products(self) -> list:
        return (await self.table("products").select("*").execute()).data

    async def fetch_subscribed_products_urls(self) -> list:
        return json.loads(
            (await self.postgres_functions().invoke("get_subscribed_products_urls", {}))["data"]
        ) or []

    async def fetch_bot_products_subscription(self, bot_deployment_id: str) -> dict:
        return (await self.table("bot_deployments").select(
            "products_subscription:products_subscriptions!product_subscription_id(id, status, desired_status)"
        ).eq(
            enums.BotDeploymentKeys.ID.value, bot_deployment_id
        ).execute()).data[0]["products_subscription"]

    async def fetch_trades(self, bot_id) -> list:
        return (await self.table("bot_trades").select("*").eq(
            enums.TradeKeys.BOT_ID.value, bot_id
        ).execute()).data

    async def reset_trades(self, bot_id):
        return (await self.table("bot_trades").delete().eq(
            enums.TradeKeys.BOT_ID.value, bot_id
        ).execute()).data

    async def upsert_trades(self, formatted_trades) -> list:
        return (await self.table("bot_trades").upsert(
            formatted_trades, on_conflict=f"{enums.TradeKeys.TRADE_ID.value},{enums.TradeKeys.TIME.value}"
        ).execute()).data

    async def update_bot_orders(self, bot_id, formatted_orders) -> dict:
        bot_update = {
            enums.BotKeys.ORDERS.value: formatted_orders
        }
        return await self.update_bot(
            bot_id, bot_update
        )

    async def fetch_bot_profile_data(self, bot_config_id: str) -> commons_profiles.ProfileData:
        if not bot_config_id:
            raise errors.MissingBotConfigError(f"bot_config_id is '{bot_config_id}'")
        bot_config = (await self.table("bot_configs").select(
            "bot_id, "
            "options, "
            "exchanges, "
            "product_config:product_configs("
            "   config, "
            "   version, "
            "   product:products!product_id(attributes)"
            ")"
        ).eq(enums.BotConfigKeys.ID.value, bot_config_id).execute()).data[0]
        profile_data = commons_profiles.ProfileData.from_dict(
            bot_config["product_config"][enums.ProfileConfigKeys.CONFIG.value]
        )
        profile_data.trading.minimal_funds = [
            commons_profiles.MinimalFund.from_dict(minimal_fund)
            for minimal_fund in bot_config["product_config"]["product"][
                enums.ProductKeys.ATTRIBUTES.value
            ].get("minimal_funds", [])
        ] if bot_config[enums.BotConfigKeys.EXCHANGES.value] else []
        profile_data.profile_details.version = bot_config["product_config"][enums.ProfileConfigKeys.VERSION.value]
        profile_data.exchanges = [
            commons_profiles.ExchangeData.from_dict(exchange_data)
            for exchange_data in bot_config[enums.BotConfigKeys.EXCHANGES.value]
        ] if bot_config[enums.BotConfigKeys.EXCHANGES.value] else []
        if options := bot_config.get(enums.BotConfigKeys.OPTIONS.value):
            profile_data.options = commons_profiles.OptionsData.from_dict(options)
        profile_data.profile_details.id = bot_config_id
        return profile_data

    async def fetch_product_config(self, product_id: str) -> commons_profiles.ProfileData:
        if not product_id:
            raise errors.MissingProductConfigError(f"product_id is '{product_id}'")
        try:
            product = (await self.table("products").select(
                "product_config:product_configs!current_config_id(config, version)"
            ).eq(enums.ProductKeys.ID.value, product_id).execute()).data[0]
        except IndexError:
            raise errors.MissingProductConfigError(f"product_id is '{product_id}'")
        profile_data = commons_profiles.ProfileData.from_dict(
            product["product_config"][enums.ProfileConfigKeys.CONFIG.value]
        )
        profile_data.profile_details.version = product["product_config"][enums.ProfileConfigKeys.VERSION.value]
        return profile_data

    async def fetch_configs(self, bot_id) -> list:
        # use a new current portfolio for the given bot
        return (await self.table("bot_configs").select("*").eq(
            enums.BotConfigKeys.BOT_ID.value, bot_id
        ).execute()).data

    async def fetch_portfolios(self, bot_id) -> list:
        return (await self.table("bot_portfolios").select("*").eq(
            enums.PortfolioKeys.BOT_ID.value, bot_id
        ).execute()).data

    async def update_portfolio(self, portfolio) -> list:
        # use a new current portfolio for the given bot
        return (await self.table("bot_portfolios").update(portfolio).eq(
            enums.PortfolioKeys.ID.value, portfolio[enums.PortfolioKeys.ID.value]
        ).execute()).data

    async def switch_portfolio(self, new_portfolio) -> dict:
        # use a new current portfolio for the given bot
        bot_id = new_portfolio[enums.PortfolioKeys.BOT_ID.value]
        inserted_portfolio = (await self.table("bot_portfolios").insert(new_portfolio).execute()).data[0]
        await self.update_bot(
            bot_id,
            {enums.BotKeys.CURRENT_PORTFOLIO_ID.value: inserted_portfolio[enums.PortfolioKeys.ID.value]}
        )
        return inserted_portfolio

    async def fetch_portfolio_history(self, portfolio_id) -> list:
        return (await self.table("bot_portfolio_histories").select("*").eq(
            enums.PortfolioHistoryKeys.PORTFOLIO_ID.value, portfolio_id
        ).execute()).data

    async def upsert_portfolio_history(self, portfolio_histories) -> list:
        return (await self.table("bot_portfolio_histories").upsert(
            portfolio_histories,
            on_conflict=f"{enums.PortfolioHistoryKeys.TIME.value},{enums.PortfolioHistoryKeys.PORTFOLIO_ID.value}"
        ).execute()).data

    async def fetch_candles_history_range(
        self, exchange: str, symbol: str, time_frame: commons_enums.TimeFrames
    ) -> (typing.Union[float, None], typing.Union[float, None]):
        min_max = json.loads(
            (await self.postgres_functions().invoke(
                "get_ohlcv_range",
                {"body": {
                    "exchange_internal_name": exchange,
                    "symbol": symbol,
                    "time_frame": time_frame.value,
                }}
            ))["data"]
        )[0]
        return (
            self.get_parsed_time(min_max["min_value"]).timestamp() if min_max["min_value"] else None,
            self.get_parsed_time(min_max["max_value"]).timestamp() if min_max["max_value"] else None,
        )

    async def fetch_candles_history(
        self, exchange: str, symbol: str, time_frame: commons_enums.TimeFrames,
        first_open_time: float, last_open_time: float
    ) -> list:
        historical_candles = await self._fetch_paginated_history(
            await self.get_production_anon_client(),
            "temp_ohlcv_history",
            "timestamp, open, high, low, close, volume",
            {
                "exchange_internal_name": exchange,
                "symbol": symbol,
                "time_frame": time_frame.value,
            },
            commons_enums.TimeFramesMinutes[time_frame] * commons_constants.MINUTE_TO_SECONDS,
            first_open_time,
            last_open_time
        )
        return self._format_ohlcvs(historical_candles)

    async def fetch_gpt_signal(
        self, exchange: str, symbol: str, time_frame: commons_enums.TimeFrames, timestamp: float, version: str
    ) -> str:
        signals = (await (await self.get_production_anon_client()).table("temp_chatgpt_signals").select("signal").match(
            {
                "timestamp": self.get_formatted_time(timestamp),
                "exchange_internal_name": exchange,
                "symbol": symbol,
                "time_frame": time_frame.value,
                "metadata->>version": version,
            },
        ).execute()).data
        if signals:
            return signals[0]["signal"]["content"]
        return ""

    async def fetch_gpt_signals_history(
        self, exchange: str, symbol: str, time_frame: commons_enums.TimeFrames,
        first_open_time: float, last_open_time: float, version: str
    ) -> dict:
        historical_signals = await self._fetch_paginated_history(
            await self.get_production_anon_client(),
            "temp_chatgpt_signals",
            "timestamp, signal",
            {
                "exchange_internal_name": exchange,
                "symbol": symbol,
                "time_frame": time_frame.value,
                "metadata->>version": version,
            },
            commons_enums.TimeFramesMinutes[time_frame] * commons_constants.MINUTE_TO_SECONDS,
            first_open_time,
            last_open_time
        )
        return self._format_gpt_signals(historical_signals)

    async def get_production_anon_client(self):
        if self.production_anon_client is None:
            self.production_anon_client = await self.init_other_postgrest_client(
                supabase_url=constants.COMMUNITY_PRODUCTION_BACKEND_URL,
                supabase_key=constants.COMMUNITY_PRODUCTION_BACKEND_KEY,
            )
        return self.production_anon_client

    async def _fetch_paginated_history(
        self, client, table_name: str, select: str, matcher: dict,
        time_interval: float, first_open_time: float, last_open_time: float
    ) -> list:
        total_elements_count = (last_open_time - first_open_time) // time_interval
        offset = 0
        max_size = 0
        total_elements = []
        max_requests_count = 100
        request_count = 0
        while request_count < max_requests_count:
            request = (
                client.table(table_name).select(select)
                .match(matcher).gte(
                    "timestamp", self.get_formatted_time(first_open_time)
                ).lte(
                    "timestamp", self.get_formatted_time(last_open_time)
                ).order(
                    "timestamp", desc=False
                )
            )
            if offset:
                request = request.range(offset, offset+max_size)
            fetched_elements = (await request.execute()).data
            total_elements += fetched_elements
            if len(fetched_elements) < max_size or (max_size == 0 and len(fetched_elements) == total_elements_count):
                # fetched everything
                break
            offset += len(fetched_elements)
            if max_size == 0:
                max_size = offset
            request_count += 1

        if request_count == max_requests_count:
            commons_logging.get_logger(self.__class__.__name__).info(
                f"paginated fetch error on {table_name} with matcher: {matcher}: "
                f"too many requests ({request_count}), fetched: {len(total_elements)} elements"
            )
        return total_elements

    def _format_gpt_signals(self, signals: list):
        return {
            self.get_parsed_time(signal["timestamp"]).timestamp(): signal["signal"]["content"]
            for signal in signals
        }

    def _format_ohlcvs(self, ohlcvs: list):
        # uses PriceIndexes order
        # IND_PRICE_TIME = 0
        # IND_PRICE_OPEN = 1
        # IND_PRICE_HIGH = 2
        # IND_PRICE_LOW = 3
        # IND_PRICE_CLOSE = 4
        # IND_PRICE_VOL = 5
        return [
            [
                int(self.get_parsed_time(ohlcv["timestamp"]).timestamp()),
                ohlcv["open"],
                ohlcv["high"],
                ohlcv["low"],
                ohlcv["close"],
                ohlcv["volume"],
            ]
            for ohlcv in ohlcvs
        ]

    async def get_asset_id(self, bucket_id: str, asset_name: str) -> str:
        """
        Not implemented for authenticated users
        """
        async with self.other_postgres_client("storage") as client:
            return (await client.from_("objects").select("*")
                .eq(
                    "bucket_id", bucket_id
                ).eq(
                    "name", asset_name
                ).execute()
            ).data[0]["id"]

    async def upload_asset(self, bucket_name: str, asset_name: str, content: typing.Union[str, bytes],) -> str:
        """
        Not implemented for authenticated users
        """
        result = await self.storage.from_(bucket_name).upload(asset_name, content)
        return asset_name

    async def send_signal(self, table, product_id: str, signal: str):
        return (await self.table(table).insert({
            enums.SignalKeys.TIME.value: self.get_formatted_time(time.time()),
            enums.SignalKeys.PRODUCT_ID.value: product_id,
            enums.SignalKeys.SIGNAL.value: signal,
        }).execute()).data[0]

    def get_subscribed_channel_tables(self) -> set:
        return set(
            channel.table_name
            for channel in self.realtime.channels
        )

    def is_realtime_connected(self) -> bool:
        return self.realtime.socket and self.realtime.socket.connected and not self.realtime.socket.closed

    @staticmethod
    def get_formatted_time(timestamp: float) -> str:
        return datetime.datetime.utcfromtimestamp(timestamp).isoformat('T')

    @staticmethod
    def get_parsed_time(str_time: str) -> datetime.datetime:
        try:
            return datetime.datetime.strptime(str_time, "%Y-%m-%dT%H:%M:%S")
        except ValueError:
            try:
                return datetime.datetime.strptime(str_time, "%Y-%m-%dT%H:%M:%S.%f")
            except ValueError:
                # last chance, try using iso format (ex: 2011-11-04T00:05:23.283+04:00)
                try:
                    return datetime.datetime.fromisoformat(str_time)
                except ValueError:
                    # sometimes fractional seconds are not supported, ex:
                    # '2023-09-04T00:01:31.06381+00:00'
                    if "." in str_time and "+" in str_time:
                        without_ms_time = str_time[0:str_time.rindex(".")] + str_time[str_time.rindex("+"):]
                        # convert to '2023-09-04T00:01:31+00:00'
                        return datetime.datetime.fromisoformat(without_ms_time)
                    raise

    async def _get_user(self) -> gotrue.User:
        return self.auth.get_user().user

    async def close(self):
        await super().close()
        if self.production_anon_client is not None:
            try:
                await self.production_anon_client.aclose()
            except RuntimeError:
                # happens when the event loop is closed already
                pass
            self.production_anon_client = None



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import pytest
import pytest_asyncio
import mock
import uuid
import json

import octobot.community as community
import octobot.community.supabase_backend.enums as supabase_enums
import octobot_commons.authentication as commons_authentication
import octobot.constants as constants
import octobot_commons.enums as commons_enums

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio

AUTH_URL = "https://oh.fake/auth"


def _build_message(value, identifier):
    return {
        commons_enums.CommunityFeedAttrs.CHANNEL_TYPE.value: commons_enums.CommunityChannelTypes.SIGNAL.value,
        commons_enums.CommunityFeedAttrs.VERSION.value: constants.COMMUNITY_FEED_CURRENT_MINIMUM_VERSION,
        commons_enums.CommunityFeedAttrs.VALUE.value: value,
        commons_enums.CommunityFeedAttrs.ID.value: identifier,
    }


@pytest_asyncio.fixture
async def authenticated_feed():
    community.IdentifiersProvider.use_production()
    authenticator = community.CommunityAuthentication(AUTH_URL, None)
    try:
        channel_mock = mock.Mock()
        channel_mock.on=mock.Mock(return_value=channel_mock)
        channel_mock.subscribe=mock.AsyncMock(return_value=channel_mock)

        realtime_mock = mock.Mock(
            channel=mock.AsyncMock(return_value=channel_mock)
        )
        authenticator.supabase_client = mock.Mock(
            sign_in=mock.AsyncMock(),
            sign_in_with_otp_token=mock.AsyncMock(),
            sign_out=mock.Mock(),
            auth=mock.Mock(_storage_key="_storage_key"),
            close=mock.AsyncMock(),
            realtime=realtime_mock,
            send_signal=mock.AsyncMock()
        )
        await authenticator._create_community_feed_if_necessary()
        yield authenticator._community_feed
    finally:
        await authenticator.stop()


async def test_start_and_connect(authenticated_feed):
    authenticated_feed.feed_callbacks = {}
    with mock.patch.object(authenticated_feed.authenticator, "is_logged_in", mock.Mock(return_value=True)) \
        as is_logged_in_mock, \
        mock.patch.object(authenticated_feed, "_subscribe_to_table", mock.AsyncMock()) \
        as _subscribe_to_table_mock:
        # without feed_callbacks
        with mock.patch.object(authenticated_feed.authenticator.supabase_client, "get_subscribed_channel_tables",
                               mock.Mock(return_value=[])) as get_subscribed_channel_tables_mock:
            await authenticated_feed.start()
            is_logged_in_mock.assert_not_called()
            _subscribe_to_table_mock.assert_not_called()
            get_subscribed_channel_tables_mock.assert_not_called()
            # with feed_callbacks
            authenticated_feed.feed_callbacks = {"signals": None, "plopplop":None}
            await authenticated_feed.start()
            assert is_logged_in_mock.call_count == 2
            # no sub channel: call _subscribe_to_table
            assert _subscribe_to_table_mock.call_count == 2
            assert get_subscribed_channel_tables_mock.call_count == 2
            get_subscribed_channel_tables_mock.reset_mock()
            is_logged_in_mock.reset_mock()
            _subscribe_to_table_mock.reset_mock()

        with mock.patch.object(authenticated_feed.authenticator.supabase_client, "get_subscribed_channel_tables",
                               mock.Mock(return_value=["signals"])) as get_subscribed_channel_tables_mock:
            # sub channel on signals: call _subscribe_to_table just for plopplop
            await authenticated_feed.start()
            is_logged_in_mock.assert_called_once()
            # no sub channel: call _subscribe_to_table
            _subscribe_to_table_mock.assert_called_with("plopplop")
            assert get_subscribed_channel_tables_mock.call_count == 2
            get_subscribed_channel_tables_mock.reset_mock()
            is_logged_in_mock.reset_mock()
            _subscribe_to_table_mock.reset_mock()

    authenticated_feed.feed_callbacks = {"signals": None, "plopplop":None}
    with pytest.raises(commons_authentication.AuthenticationRequired):
        with mock.patch.object(authenticated_feed.authenticator.supabase_client, "get_subscribed_channel_tables",
                               mock.Mock(return_value=["plop"])), \
             mock.patch.object(authenticated_feed.authenticator, "is_logged_in", mock.Mock(return_value=False)):
                await authenticated_feed.start()

async def test_connection(authenticated_feed):
    with mock.patch.object(authenticated_feed.authenticator, "is_logged_in", mock.Mock(return_value=True)) \
         as is_logged_in_mock:
        with mock.patch.object(authenticated_feed.authenticator.supabase_client, "is_realtime_connected",
                               mock.Mock(return_value=True)) as is_realtime_connected_mock:
            assert authenticated_feed.is_connected_to_remote_feed() is True
            is_realtime_connected_mock.assert_called_once()
            assert authenticated_feed.is_connected() is True
            is_logged_in_mock.assert_called_once()
    with mock.patch.object(authenticated_feed.authenticator, "is_logged_in", mock.Mock(return_value=False)) \
        as is_logged_in_mock:
        with mock.patch.object(authenticated_feed.authenticator.supabase_client, "is_realtime_connected",
                               mock.Mock(return_value=False)) as is_realtime_connected_mock:
            assert authenticated_feed.is_connected_to_remote_feed() is False
            is_realtime_connected_mock.assert_called_once()
            assert authenticated_feed.is_connected() is False
            is_logged_in_mock.assert_called_once()


async def test_subscribe_to_table(authenticated_feed):
    channel_mock = await authenticated_feed._realtime_client.channel()
    authenticated_feed._realtime_client.channel.reset_mock()
    channel_mock.on.assert_not_called()
    channel_mock.subscribe.assert_not_called()

    await authenticated_feed._subscribe_to_table("plop_table")
    authenticated_feed._realtime_client.channel.assert_called_once_with(
        authenticated_feed.SCHEMA, "plop_table"
    )
    channel_mock.on.assert_called_once_with(
        authenticated_feed.INSERT_EVENT, authenticated_feed._message_callback
    )
    channel_mock.subscribe.assert_called_once_with(
        authenticated_feed._subscribed_callback
    )


async def test_register_feed_callback(authenticated_feed):
    assert authenticated_feed.is_signal_receiver is False
    assert authenticated_feed.feed_callbacks == {}
    target_table = authenticated_feed._get_table(commons_enums.CommunityChannelTypes.SIGNAL)

    with mock.patch.object(authenticated_feed, "_subscribe_to_table_if_necessary", mock.AsyncMock()) \
        as _subscribe_to_table_if_necessary_mock:
        await authenticated_feed.register_feed_callback(
            commons_enums.CommunityChannelTypes.SIGNAL, "plop_cb", "identifier"
        )
        assert authenticated_feed.is_signal_receiver is True
        assert authenticated_feed.feed_callbacks == {
            target_table: {
                "identifier": ["plop_cb"]
            }
        }
        _subscribe_to_table_if_necessary_mock.assert_called_once_with(target_table)
        _subscribe_to_table_if_necessary_mock.reset_mock()

        await authenticated_feed.register_feed_callback(
            commons_enums.CommunityChannelTypes.SIGNAL, "plop_cb", "identifier"
        )
        assert authenticated_feed.feed_callbacks == {
            target_table: {
                "identifier": ["plop_cb", "plop_cb"]    # added callback
            }
        }
        _subscribe_to_table_if_necessary_mock.assert_called_once_with(target_table)
        _subscribe_to_table_if_necessary_mock.reset_mock()

        await authenticated_feed.register_feed_callback(
            commons_enums.CommunityChannelTypes.SIGNAL, "plop_cb2"
        )
        assert authenticated_feed.feed_callbacks == {
            target_table: {
                "identifier": ["plop_cb", "plop_cb"],
                None: ["plop_cb2"]
            }
        }
        _subscribe_to_table_if_necessary_mock.assert_called_once_with(target_table)
        _subscribe_to_table_if_necessary_mock.reset_mock()

        # wrong table
        with pytest.raises(NotImplementedError):
            await authenticated_feed.register_feed_callback(
                "wrong type", "plop_cb3"
            )
        assert authenticated_feed.feed_callbacks == {
            target_table: {
                "identifier": ["plop_cb", "plop_cb"],
                None: ["plop_cb2"]
            }
        }
        _subscribe_to_table_if_necessary_mock.assert_not_called()


async def test_send(authenticated_feed):
    assert authenticated_feed.is_signal_receiver is False
    assert authenticated_feed.is_signal_emitter is False
    message = {"text": "ploppp"}
    channel_type = commons_enums.CommunityChannelTypes.SIGNAL
    identifier = "plop_identifier"

    with mock.patch.object(authenticated_feed.authenticator, "is_logged_in", mock.Mock(return_value=False)) \
        as is_logged_in_mock:
        await authenticated_feed.send(message, channel_type, identifier)
        is_logged_in_mock.assert_called_once()

    with mock.patch.object(authenticated_feed.authenticator, "is_logged_in", mock.Mock(return_value=True)) \
        as is_logged_in_mock, \
            mock.patch.object(uuid, "uuid4", mock.Mock(return_value="1")) as uuid4_mock:
        await authenticated_feed.send(message, channel_type, identifier)
        is_logged_in_mock.assert_called_once()
        authenticated_feed.authenticator.supabase_client.send_signal.assert_called_once_with(
            authenticated_feed.SIGNALS_TABLE,
            identifier,
            authenticated_feed._build_message(channel_type, message)
        )
        assert uuid4_mock.call_count == 2   # 2 calls to _build_message


async def test_message_callback(authenticated_feed):
    message = {"text": "ploppp"}
    channel_type = commons_enums.CommunityChannelTypes.SIGNAL
    identifier = "plop_identifier"
    message_dict = {
        "table": authenticated_feed._get_table(channel_type),
        "new": {
            "signal": json.loads(authenticated_feed._build_message(channel_type, message)),
            supabase_enums.SignalKeys.PRODUCT_ID.value: identifier
        },
    }
    # no callback
    await authenticated_feed._message_callback(message_dict)
    callback = mock.AsyncMock()
    with mock.patch.object(authenticated_feed, "_subscribe_to_table_if_necessary", mock.AsyncMock()) \
        as _subscribe_to_table_if_necessary_mock:
        await authenticated_feed.register_feed_callback(channel_type, callback, identifier=identifier)
        _subscribe_to_table_if_necessary_mock.assert_called_once()

    # new message for new message_id
    json_message = json.loads(authenticated_feed._build_message(channel_type, message))
    message_dict["new"]["signal"] = json_message
    await authenticated_feed._message_callback(message_dict)
    callback.assert_called_once_with(json_message)
    callback.reset_mock()
    # same message, ignored
    await authenticated_feed._message_callback(message_dict)
    callback.assert_not_called()

    # register other callbacks
    callback_2 = mock.AsyncMock()
    callback_3 = mock.AsyncMock()
    callback_4 = mock.AsyncMock()
    with mock.patch.object(authenticated_feed, "_subscribe_to_table_if_necessary", mock.AsyncMock()) \
        as _subscribe_to_table_if_necessary_mock:
        await authenticated_feed.register_feed_callback(channel_type, callback_2, identifier=identifier)
        await authenticated_feed.register_feed_callback(channel_type, callback_3, identifier="other")
        await authenticated_feed.register_feed_callback(channel_type, callback_4)
        assert _subscribe_to_table_if_necessary_mock.call_count == 3

    # new message for new message_id
    json_message = json.loads(authenticated_feed._build_message(channel_type, message))
    message_dict["new"]["signal"] = json_message
    await authenticated_feed._message_callback(message_dict)
    callback.assert_called_once_with(json_message)
    callback_2.assert_called_once_with(json_message)
    callback_3.assert_not_called()    # different identifier
    callback_4.assert_not_called()    # different identifier



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.


_INNER_BOT_QUERY = """
    _id
    config {
        _id
        profile {
            _id
            name
        }
    }
    name
    device {
      _id
      host
      name
      region
      uuid
    }
    deployment {
      _id
      type
      urls {
        url
      }
    }
    stats {
        _id
        profitability
    }
    user_id
"""


def select_startup_info_query(bot_id) -> (str, dict, str):
    return """
query getBotStartupInfo($bot_id: String) {
  getBotStartupInfo(input: {botId: $bot_id}) {
    subscribedProducts {
        url
    }
    forcedProfileUrl {
        url
    }
  }
}
    """, {"bot_id": bot_id}, "getBotStartupInfo"


def select_subscribed_profiles_query() -> (str, dict, str):
    return """
query getSubscribedProfiles {
  getSubscribedProfiles {
    data {
        url
    }
  }
}
    """, {}, "getSubscribedProfiles"


def select_bots_query() -> (str, dict, str):
    return """
query SelectBots {  
    bots {
""" + _INNER_BOT_QUERY + """
    }
}
    """, {}, "bots"


def select_bot_query(bot_id) -> (str, dict, str):
    return """
query SelectBot($_id: ObjectId) {  
    bot (query: {_id: $_id}) {
""" + _INNER_BOT_QUERY + """
    }
}
    """, {"_id": bot_id},  "bot"


def create_bot_query(is_self_hosted) -> (str, dict, str):
    return """
mutation CreateBot($isSelfHosted: Boolean) {  
    createBot (input: {isSelfHosted: $isSelfHosted}) {
""" + _INNER_BOT_QUERY + """
    }
}
    """, {"isSelfHosted": is_self_hosted}, "createBot"


def create_bot_device_query(bot_id) -> (str, dict, str):
    return """
mutation CreateBotDevice($bot_id: ObjectId) {
  createBotDevice(input: $bot_id) {
    """ + _INNER_BOT_QUERY + """
  }
}
    """, {"bot_id": bot_id}, "createBotDevice"


def update_bot_config_and_stats_query(bot_id, profile_name, profitability) -> (str, dict, str):
    return """
mutation updateOneBot($bot_id: ObjectId, $profile_name: String, $profitability: Decimal) {
  updateOneBot(
    query: {_id: $bot_id}
    set: {config: {profile: {name: $profile_name}}, stats: {profitability: $profitability}}
  ) {
    """ + _INNER_BOT_QUERY + """
  }
}
    """, {"bot_id": bot_id, "profile_name": profile_name, "profitability": str(profitability)}, "updateOneBot"


def update_bot_trades_query(bot_id, trades) -> (str, dict, str):
    return """
mutation updateOneBot($bot_id: ObjectId, $trades: [BotTradeUpdateInput]) {
  updateOneBot(
    query: {_id: $bot_id}
    set: {trades: $trades}
  ) {
    """ + _INNER_BOT_QUERY + """
  }
}
    """, {"bot_id": bot_id, "trades": trades}, "updateOneBot"


def upsert_bot_trades_query(bot_id, trades) -> (str, dict, str):
    return """
mutation upsertBotTrades($bot_id: String, $trades: [BotTradesUpsertInputTrade]) {
  upsertBotTrades(input: {bot_id: $bot_id, trades: $trades}){
    status
  }
}
    """, {"bot_id": bot_id, "trades": trades}, "upsertBotTrades"


def update_bot_portfolio_query(bot_id, current_value, initial_value, unit, content, history) -> (str, dict, str):
    return """
mutation updateOneBot($bot_id: ObjectId, $current_value: Decimal, $initial_value: Decimal, $unit: String, $content: [BotPortfolioContentUpdateInput], $history: [BotPortfolioHistoryUpdateInput]) {
  updateOneBot(
    query: {_id: $bot_id}
    set: {portfolio: {current_value: $current_value, initial_value: $initial_value, unit: $unit, content: $content, history: $history}}
  ) {
    """ + _INNER_BOT_QUERY + """
  }
}
    """, {"bot_id": bot_id, "current_value": str(current_value), "initial_value": str(initial_value),
          "unit": unit, "content": content, "history": history}, "updateOneBot"


def upsert_historical_bot_portfolio_query(bot_id, current_value, content, history) -> (str, dict, str):
    return """
mutation upsertBotPortfolio($bot_id: String, $current_value: Decimal, $content: [BotPortfolioUpsertInputContent], $history: [BotPortfolioUpsertInputHistory]) {
  upsertBotPortfolio(input: {bot_id: $bot_id, current_value: $current_value, content: $content, history: $history}){
    status
  }
}
    """, {"bot_id": bot_id, "current_value": str(current_value), "content": content, "history": history}, "upsertBotPortfolio"




#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
from copy import deepcopy

from octobot_backtesting.importers.exchanges.exchange_importer import ExchangeDataImporter
from octobot_commons.enums import TimeFrames
from octobot_commons.enums import PriceIndexes
import octobot_trading.exchange_data as exchange_data
from octobot_trading.api.symbol_data import create_new_candles_manager
from octobot_trading.exchanges.exchange_manager import ExchangeManager
from octobot_trading.util.initializable import Initializable

"""
Class containing data with known moves
Moves are in the middle of the return data_frame

TimeFrames.FIVE_MINUTES: 2018-04-26 17:00:00 -> 2018-04-27 02:00:00
TimeFrames.FIFTEEN_MINUTES: 2018-04-23 06:15:00 -> 2018-04-24 07:00:00
TimeFrames.ONE_HOUR: 2018-04-07 16:00:00 -> 2018-04-11 20:00:00
TimeFrames.FOUR_HOURS: 2018-02-03 00:00:00 -> 2018-02-20 16:00:00
TimeFrames.HEIGHT_HOURS: 2017-11-12 04:00:00 -> 2017-12-11 16:00:00
TimeFrames.ONE_DAY: 2017-08-17 00:00:00 -> 2017-11-12 00:00:00

"""


class DataBank(Initializable):

    def __init__(self, data_file=None):
        super().__init__()
        self.data_file = data_file if data_file else "tests/static/AbstractExchangeHistoryCollector_1586017993.616272.data"
        self.exchange_name = None
        self.data_importer = ExchangeDataImporter({}, self.data_file)
        self.default_symbol = "BTC/USDT"

        self.origin_ohlcv_by_symbol = {}
        self.candles_managers_by_time_frame = {}
        self.current_init_indexes_by_time_frame = {}
        self.current_data = None
        self.data_by_symbol_by_data_frame = None
        self.manager = ExchangeManager({}, "binance")
        self.symbol_data = exchange_data.ExchangeSymbolData(self.manager, self.default_symbol)
        self.symbol_data.symbol_candles = self.candles_managers_by_time_frame

    async def initialize_impl(self):
        await self.data_importer.initialize()
        self.exchange_name = self.data_importer.exchange_name
        for time_frame in self.data_importer.time_frames:
            for symbol in self.data_importer.symbols:
                if symbol not in self.origin_ohlcv_by_symbol:
                    self.origin_ohlcv_by_symbol[symbol] = {}
                db_data = await self.data_importer.get_ohlcv_from_timestamps(
                    exchange_name=self.exchange_name,
                    symbol=symbol,
                    time_frame=time_frame
                )
                # store ohlcv only
                self.origin_ohlcv_by_symbol[symbol][time_frame] = sorted(
                    [data[-1] for data in db_data],
                    key=lambda x: x[PriceIndexes.IND_PRICE_TIME.value]
                )

    async def stop(self):
        await self.data_importer.stop()
        await self.manager.stop()

    def get_time_frames(self):
        return self.data_importer.time_frames

    def get_last_candle_for_default_symbol(self, time_frame):
        return [val[0] for val in self.candles_managers_by_time_frame[time_frame].get_symbol_prices(1).values()]

    def set_data_for_default_symbol(self, time_frame, end_index):
        self._set_candle_manager(time_frame, end_index, use_current_data=True)

    def increment_data_for_default_symbol(self, time_frame):
        current_index = self.current_init_indexes_by_time_frame[time_frame]
        candle = self.origin_ohlcv_by_symbol[self.default_symbol][time_frame][current_index]
        self.candles_managers_by_time_frame[time_frame].add_new_candle(candle)
        self.current_init_indexes_by_time_frame[time_frame] += 1

    # use default data with full data range
    def standard_mode(self, initial_candles):
        for time_frame in self.get_time_frames():
            self._set_candle_manager(time_frame, initial_candles)

    # works only with default data file
    # not started, selling started, selling maxed, buying starting, max: back normal:
    def rise_after_over_sold_mode(self):
        self.current_data = self._merge_candles(
            self.origin_ohlcv_by_symbol[self.default_symbol][TimeFrames.FOUR_HOURS][35:61],
            self.origin_ohlcv_by_symbol[self.default_symbol][TimeFrames.FOUR_HOURS][0:49]
        )
        return TimeFrames.FOUR_HOURS, 35, 43, 46, 47, 75

    # works only with default data file
    # not started, buying started, buying maxed, start dipping, super dip, max: back normal:
    def dip_after_over_bought_mode(self):
        self.current_data = self.origin_ohlcv_by_symbol[self.default_symbol][TimeFrames.ONE_DAY]
        return TimeFrames.ONE_DAY, 72, 76, 80, 81, 88, 90

    # works only with default data file
    # not started, started, heavy dump, very light dump, max: stopped dump:
    def sudden_dump_mode(self):
        self.current_data = self.origin_ohlcv_by_symbol[self.default_symbol][TimeFrames.ONE_HOUR]
        return TimeFrames.ONE_HOUR, 42, 43, 44, 45, 46

    # works only with default data file
    # not started, started, heavy pump, max pump, change trend, dipping, max: dipped:
    def sudden_pump_mode(self):
        self.current_data = self.origin_ohlcv_by_symbol[self.default_symbol][TimeFrames.HEIGHT_HOURS]
        return TimeFrames.HEIGHT_HOURS, 71, 74, 76, 77, 78, 79, 83

    # works only with default data file
    # long data_frame with flat then sudden big rise and then mostly flat for 120 values
    # start move ending up in a rise, reaches flat trend, first micro up p1, first mirco up p2, micro down, micro up,
    # micro down, micro up, micro down, back normal, micro down, back normal, micro down, back up, micro up, back down,
    # back normal, micro down, back up, micro down, back up, micro down, back up
    def overall_flat_trend_mode(self):
        cheap_prices = [
            [data * 0.9 for data in candle]
            for candle in self.origin_ohlcv_by_symbol[self.default_symbol][TimeFrames.FIFTEEN_MINUTES][6:72]
        ]
        self.current_data = self._merge_candles(
            cheap_prices,
            self.origin_ohlcv_by_symbol[self.default_symbol][TimeFrames.HEIGHT_HOURS][25:43],
            self.origin_ohlcv_by_symbol[self.default_symbol][TimeFrames.FIFTEEN_MINUTES][6:72],
            self.origin_ohlcv_by_symbol[self.default_symbol][TimeFrames.FIFTEEN_MINUTES][6:72]
        )
        return TimeFrames.FIFTEEN_MINUTES, 67, 85, 87, 90, 95, 107, 116, 123, 138, 141, 151, 153, 161, 163, 173, \
            174, 175, 180, 183, 193, 198, 203, 207

    def _set_candle_manager(self, time_frame, initial_candles_max_index, use_current_data=False):
        self.candles_managers_by_time_frame[time_frame] = create_new_candles_manager(
            self.current_data[:initial_candles_max_index] if use_current_data else
            self.origin_ohlcv_by_symbol[self.default_symbol][time_frame][:initial_candles_max_index]
        )
        self.current_init_indexes_by_time_frame[time_frame] = initial_candles_max_index

    @staticmethod
    def _merge_candles(*candles_list):
        result_list = deepcopy(candles_list[0])
        timestamp_list = [candle[PriceIndexes.IND_PRICE_TIME.value] for candle in result_list]
        for candle_list in candles_list[1:]:
            # ensure no timestamp is present in 2 candles
            while any(candle[PriceIndexes.IND_PRICE_TIME.value] in timestamp_list for candle in candle_list):
                for candle in candle_list:
                    candle[PriceIndexes.IND_PRICE_TIME.value] = candle[PriceIndexes.IND_PRICE_TIME.value] + 1
            result_list += deepcopy(candle_list)
            timestamp_list = [candle[PriceIndexes.IND_PRICE_TIME.value] for candle in result_list]
        return result_list



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import octobot.backtesting as backtesting
import octobot_backtesting.constants as constants


def create_independent_backtesting(
    config,
    tentacles_setup_config,
    data_files,
    data_file_path=constants.BACKTESTING_FILE_PATH,
    join_backtesting_timeout=constants.BACKTESTING_DEFAULT_JOIN_TIMEOUT,
    run_on_common_part_only=True,
    start_timestamp=None,
    end_timestamp=None,
    enable_logs=True,
    stop_when_finished=False,
    name=None,
    enforce_total_databases_max_size_after_run=True,
    enable_storage=True,
    run_on_all_available_time_frames=False,
    backtesting_data=None,
    config_by_tentacle=None,
    services_config=None
) -> backtesting.IndependentBacktesting:
    return backtesting.IndependentBacktesting(
        config, tentacles_setup_config, data_files,
        data_file_path,
        run_on_common_part_only=run_on_common_part_only,
        join_backtesting_timeout=join_backtesting_timeout,
        start_timestamp=start_timestamp,
        end_timestamp=end_timestamp,
        enable_logs=enable_logs,
        stop_when_finished=stop_when_finished,
        name=name,
        enforce_total_databases_max_size_after_run=enforce_total_databases_max_size_after_run,
        enable_storage=enable_storage,
        run_on_all_available_time_frames=run_on_all_available_time_frames,
        backtesting_data=backtesting_data,
        config_by_tentacle=config_by_tentacle,
        services_config=services_config,
    )


async def initialize_and_run_independent_backtesting(independent_backtesting, log_errors=True) -> None:
    await independent_backtesting.initialize_and_run(log_errors=log_errors)


async def join_independent_backtesting(independent_backtesting, timeout=constants.BACKTESTING_DEFAULT_JOIN_TIMEOUT) -> None:
    await independent_backtesting.join_backtesting_updater(timeout)


async def initialize_independent_backtesting_config(independent_backtesting) -> dict:
    return await independent_backtesting.initialize_config()


async def clear_backtesting_fetched_data(independent_backtesting):
    await independent_backtesting.clear_fetched_data()


async def stop_independent_backtesting(independent_backtesting, memory_check=False, should_raise=False) -> None:
    await independent_backtesting.stop(memory_check=memory_check, should_raise=should_raise)


async def join_independent_backtesting_stop(independent_backtesting, timeout=None):
    await independent_backtesting.join_stop_event(timeout=timeout)


def check_independent_backtesting_remaining_objects(independent_backtesting) -> None:
    independent_backtesting.octobot_backtesting.check_remaining_objects()


def is_independent_backtesting_in_progress(independent_backtesting) -> bool:
    return independent_backtesting.is_in_progress()


def is_independent_backtesting_computing(independent_backtesting) -> bool:
    return independent_backtesting.is_in_progress()


def get_independent_backtesting_progress(independent_backtesting) -> float:
    return independent_backtesting.get_progress()


def is_independent_backtesting_finished(independent_backtesting) -> bool:
    return independent_backtesting.has_finished()


def is_independent_backtesting_stopped(independent_backtesting) -> bool:
    return independent_backtesting.stopped


async def get_independent_backtesting_report(independent_backtesting) -> dict:
    return await independent_backtesting.get_dict_formatted_report()


def get_independent_backtesting_exchange_manager_ids(independent_backtesting) -> list:
    return independent_backtesting.octobot_backtesting.exchange_manager_ids


def get_independent_backtesting_bot_id(independent_backtesting) -> str:
    return independent_backtesting.octobot_backtesting.bot_id


def log_independent_backtesting_report(independent_backtesting) -> None:
    independent_backtesting.log_report()


def get_independent_backtesting_config(independent_backtesting) -> dict:
    return independent_backtesting.octobot_backtesting.backtesting_config


def get_independent_backtesting_symbols_by_exchanges(independent_backtesting) -> dict:
    return independent_backtesting.octobot_backtesting.symbols_to_create_exchange_classes



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.

from octobot.backtesting import abstract_backtesting_test
from octobot.backtesting import independent_backtesting
from octobot.backtesting import octobot_backtesting
from octobot.backtesting.abstract_backtesting_test import (
    AbstractBacktestingTest,
)
from octobot.backtesting.independent_backtesting import (
    IndependentBacktesting,
)
from octobot.backtesting.octobot_backtesting import (
    OctoBotBacktesting,
)

__all__ = [
    "OctoBotBacktesting",
    "IndependentBacktesting",
    "AbstractBacktestingTest",
]
